"""
Canned response for monitoring json home
"""


def json_home(url):
    """
    Canned response for the json_home call
    """
    return {
        "resources": {
            url + "/{tenantId}/account": {
                "href-template": "/v1.0/{tenantId}/account",
                "href-vars": {
                    "tenantId": "http://docs.rackspace.com/cm/api/v1.0/cm-devguide/params/tenantId"
                },
                "hints": {
                    "docs": ("http://docs.rackspace.com/cm/api/v1.0/cm-devguide"
                             "/content/service-account.html"),
                    "allow": [
                        "GET",
                        "PUT"
                    ],
                    "representations": [
                        "application/json",
                        "application/xml"
                    ],
                    "accept-put": [
                        "application/xml",
                        "application/json"
                    ]
                }
            },
            url + "/{tenantId}/agents": {
                "href-template": "/v1.0/{tenantId}/agents",
                "href-vars": {
                    "tenantId": "http://docs.rackspace.com/cm/api/v1.0/cm-devguide/params/tenantId"
                },
                "hints": {
                    "docs": ("http://docs.rackspace.com/cm/api/v1.0/cm-devguide"
                             "/content/service-agents.html"),
                    "allow": [
                        "GET"
                    ],
                    "representations": [
                        "application/json",
                        "application/xml"
                    ]
                }
            },
            url + "/{tenantId}/agents/{agentId}": {
                "href-template": "/v1.0/{tenantId}/agents/{agentId}",
                "href-vars": {
                    "tenantId": "http://docs.rackspace.com/cm/api/v1.0/cm-devguide/params/tenantId",
                    "agentId": "http://docs.rackspace.com/cm/api/v1.0/cm-devguide/params/agentId"
                },
                "hints": {
                    "docs": ("http://docs.rackspace.com/cm/api/v1.0/cm-devguide"
                             "/content/service-agents.html"),
                    "allow": [
                        "GET"
                    ],
                    "representations": [
                        "application/json",
                        "application/xml"
                    ]
                }
            },
            url + "/{tenantId}/agents/{agentId}/connections": {
                "href-template": "/v1.0/{tenantId}/agents/{agentId}/connections",
                "href-vars": {
                    "tenantId": "http://docs.rackspace.com/cm/api/v1.0/cm-devguide/params/tenantId",
                    "agentId": "http://docs.rackspace.com/cm/api/v1.0/cm-devguide/params/agentId"
                },
                "hints": {
                    "docs": ("http://docs.rackspace.com/cm/api/v1.0/cm-devguide/content/"
                             "service-agent-connections.html"),
                    "allow": [
                        "GET"
                    ],
                    "representations": [
                        "application/json",
                        "application/xml"
                    ]
                }
            },
            url + "/{tenantId}/agents/{agentId}/connections/{connId}": {
                "href-template": "/v1.0/{tenantId}/agents/{agentId}/connections/{connId}",
                "href-vars": {
                    "tenantId": "http://docs.rackspace.com/cm/api/v1.0/cm-devguide/params/tenantId",
                    "agentId": "http://docs.rackspace.com/cm/api/v1.0/cm-devguide/params/agentId",
                    "connId": "http://docs.rackspace.com/cm/api/v1.0/cm-devguide/params/connId"
                },
                "hints": {
                    "docs": ("http://docs.rackspace.com/cm/api/v1.0/cm-devguide"
                             "/content/service-agent-connections.html"),
                    "allow": [
                        "GET"
                    ],
                    "representations": [
                        "application/json",
                        "application/xml"
                    ]
                }
            },
            url + "/{tenantId}/agents/{agentId}/host_info_types": {
                "href-template": "/v1.0/{tenantId}/agents/{agentId}/host_info_types",
                "href-vars": {
                    "tenantId": "http://docs.rackspace.com/cm/api/v1.0/cm-devguide/params/tenantId",
                    "agentId": "http://docs.rackspace.com/cm/api/v1.0/cm-devguide/params/agentId"
                },
                "hints": {
                    "docs": ("http://docs.rackspace.com/cm/api/v1.0/cm-devguide"
                             "/content/service-agents.html"),
                    "allow": [
                        "GET"
                    ],
                    "representations": [
                        "application/json",
                        "application/xml"
                    ]
                }
            },
            url + "/{tenantId}/agents/{agentId}/host_info/{handlerType}": {
                "href-template": "/v1.0/{tenantId}/agents/{agentId}/host_info/{handlerType}",
                "href-vars": {
                    "tenantId": "http://docs.rackspace.com/cm/api/v1.0/cm-devguide/params/tenantId",
                    "agentId": "http://docs.rackspace.com/cm/api/v1.0/cm-devguide/params/agentId",
                    "handlerType": "http://docs.rackspace.com/cm/api/v1.0/cm-devguide/params/handlerType"
                },
                "hints": {
                    "docs": ("http://docs.rackspace.com/cm/api/v1.0/cm-devguide"
                             "/content/service-agents.html"),
                    "allow": [
                        "GET"
                    ],
                    "representations": [
                        "application/json",
                        "application/xml"
                    ]
                }
            },
            url + "/{tenantId}/agent_tokens": {
                "href-template": "/v1.0/{tenantId}/agent_tokens",
                "href-vars": {
                    "tenantId": "http://docs.rackspace.com/cm/api/v1.0/cm-devguide/params/tenantId"
                },
                "hints": {
                    "docs": ("http://docs.rackspace.com/cm/api/v1.0/cm-devguide"
                             "/content/service-agent-tokens.html"),
                    "allow": [
                        "GET",
                        "POST"
                    ],
                    "representations": [
                        "application/json",
                        "application/xml"
                    ],
                    "accept-post": [
                        "application/xml",
                        "application/json"
                    ]
                }
            },
            url + "/{tenantId}/agent_tokens/{tokenId}": {
                "href-template": "/v1.0/{tenantId}/agent_tokens/{tokenId}",
                "href-vars": {
                    "tenantId": "http://docs.rackspace.com/cm/api/v1.0/cm-devguide/params/tenantId",
                    "tokenId": "http://docs.rackspace.com/cm/api/v1.0/cm-devguide/params/tokenId"
                },
                "hints": {
                    "docs": ("http://docs.rackspace.com/cm/api/v1.0/cm-devguide"
                             "/content/service-agent-tokens.html"),
                    "allow": [
                        "GET",
                        "PUT",
                        "DELETE"
                    ],
                    "representations": [
                        "application/json",
                        "application/xml"
                    ],
                    "accept-put": [
                        "application/xml",
                        "application/json"
                    ]
                }
            },
            url + "/{tenantId}/agent_installers": {
                "href-template": "/v1.0/{tenantId}/agent_installers",
                "href-vars": {
                    "tenantId": "http://docs.rackspace.com/cm/api/v1.0/cm-devguide/params/tenantId"
                },
                "hints": {
                    "docs": ("http://docs.rackspace.com/cm/api/v1.0/cm-devguide"
                             "/content/service-agent-installers.html"),
                    "allow": [
                        "POST"
                    ],
                    "representations": [
                        "application/json",
                        "application/xml"
                    ],
                    "accept-post": [
                        "application/xml",
                        "application/json"
                    ]
                }
            },
            url + "/{tenantId}/agent_installers/{installerId.sh}": {
                "href-template": "/v1.0/{tenantId}/agent_installers/{installerId.sh}",
                "href-vars": {
                    "tenantId": "http://docs.rackspace.com/cm/api/v1.0/cm-devguide/params/tenantId",
                    "installerId.sh": ("http://docs.rackspace.com/cm/api/v1.0/cm-devguide"
                                       "/params/installerId.sh")
                },
                "hints": {
                    "docs": ("http://docs.rackspace.com/cm/api/v1.0/cm-devguide"
                             "/content/service-agent-installers.html"),
                    "allow": [
                        "GET"
                    ],
                    "representations": [
                        "application/json",
                        "application/xml"
                    ]
                }
            },
            url + "/{tenantId}/agent_installers/{installerId}": {
                "href-template": "/v1.0/{tenantId}/agent_installers/{installerId}",
                "href-vars": {
                    "tenantId": "http://docs.rackspace.com/cm/api/v1.0/cm-devguide/params/tenantId",
                    "installerId": "http://docs.rackspace.com/cm/api/v1.0/cm-devguide/params/installerId"
                },
                "hints": {
                    "docs": ("http://docs.rackspace.com/cm/api/v1.0/cm-devguide"
                             "/content/service-agent-installers.html"),
                    "allow": [
                        "GET"
                    ],
                    "representations": [
                        "application/json",
                        "application/xml"
                    ]
                }
            },
            url + "/{tenantId}/monitoring_zones": {
                "href-template": "/v1.0/{tenantId}/monitoring_zones",
                "href-vars": {
                    "tenantId": "http://docs.rackspace.com/cm/api/v1.0/cm-devguide/params/tenantId"
                },
                "hints": {
                    "docs": ("http://docs.rackspace.com/cm/api/v1.0/cm-devguide"
                             "/content/service-monitoring-zones.html"),
                    "allow": [
                        "GET"
                    ],
                    "representations": [
                        "application/json",
                        "application/xml"
                    ]
                }
            },
            url + "/{tenantId}/monitoring_zones/{monitoringZoneId}": {
                "href-template": "/v1.0/{tenantId}/monitoring_zones/{monitoringZoneId}",
                "href-vars": {
                    "tenantId": "http://docs.rackspace.com/cm/api/v1.0/cm-devguide/params/tenantId",
                    "monitoringZoneId": ("http://docs.rackspace.com/cm/api/v1.0/"
                                         "cm-devguide/params/monitoringZoneId")
                },
                "hints": {
                    "docs": ("http://docs.rackspace.com/cm/api/v1.0/cm-devguide"
                             "/content/service-monitoring-zones.html"),
                    "allow": [
                        "GET"
                    ],
                    "representations": [
                        "application/json",
                        "application/xml"
                    ]
                }
            },
            url + "/{tenantId}/monitoring_zones/{monitoringZoneId}/traceroute": {
                "href-template": "/v1.0/{tenantId}/monitoring_zones/{monitoringZoneId}/traceroute",
                "href-vars": {
                    "tenantId": "http://docs.rackspace.com/cm/api/v1.0/cm-devguide/params/tenantId",
                    "monitoringZoneId": ("http://docs.rackspace.com/cm/api/v1.0/cm-devguide"
                                         "/params/monitoringZoneId")
                },
                "hints": {
                    "docs": ("http://docs.rackspace.com/cm/api/v1.0/cm-devguide"
                             "/content/service-monitoring-zones.html"),
                    "allow": [
                        "POST"
                    ],
                    "representations": [
                        "application/json",
                        "application/xml"
                    ],
                    "accept-post": [
                        "application/xml",
                        "application/json"
                    ]
                }
            },
            url + "/{tenantId}/changelogs/alarms": {
                "href-template": "/v1.0/{tenantId}/changelogs/alarms",
                "href-vars": {
                    "tenantId": "http://docs.rackspace.com/cm/api/v1.0/cm-devguide/params/tenantId"
                },
                "hints": {
                    "docs": ("http://docs.rackspace.com/cm/api/v1.0/cm-devguide"
                             "/content/service-alarms.html"),
                    "allow": [
                        "GET"
                    ],
                    "representations": [
                        "application/json",
                        "application/xml"
                    ]
                }
            },
            url + "/{tenantId}/entities": {
                "href-template": "/v1.0/{tenantId}/entities",
                "href-vars": {
                    "tenantId": "http://docs.rackspace.com/cm/api/v1.0/cm-devguide/params/tenantId"
                },
                "hints": {
                    "docs": ("http://docs.rackspace.com/cm/api/v1.0/cm-devguide"
                             "/content/service-entities.html"),
                    "allow": [
                        "GET",
                        "POST"
                    ],
                    "representations": [
                        "application/json",
                        "application/xml"
                    ],
                    "accept-post": [
                        "application/xml",
                        "application/json"
                    ]
                }
            },
            url + "/{tenantId}/entities/{entityId}": {
                "href-template": "/v1.0/{tenantId}/entities/{entityId}",
                "href-vars": {
                    "tenantId": "http://docs.rackspace.com/cm/api/v1.0/cm-devguide/params/tenantId",
                    "entityId": "http://docs.rackspace.com/cm/api/v1.0/cm-devguide/params/entityId"
                },
                "hints": {
                    "docs": ("http://docs.rackspace.com/cm/api/v1.0/cm-devguide"
                             "/content/service-entities.html"),
                    "allow": [
                        "GET",
                        "PUT",
                        "DELETE"
                    ],
                    "representations": [
                        "application/json",
                        "application/xml"
                    ],
                    "accept-put": [
                        "application/xml",
                        "application/json"
                    ]
                }
            },
            url + "/{tenantId}/entities/{entityId}/checks": {
                "href-template": "/v1.0/{tenantId}/entities/{entityId}/checks",
                "href-vars": {
                    "tenantId": "http://docs.rackspace.com/cm/api/v1.0/cm-devguide/params/tenantId",
                    "entityId": "http://docs.rackspace.com/cm/api/v1.0/cm-devguide/params/entityId"
                },
                "hints": {
                    "docs": ("http://docs.rackspace.com/cm/api/v1.0/cm-devguide"
                             "/content/service-checks.html"),
                    "allow": [
                        "GET",
                        "POST"
                    ],
                    "representations": [
                        "application/json",
                        "application/xml"
                    ],
                    "accept-post": [
                        "application/xml",
                        "application/json"
                    ]
                }
            },
            url + "/{tenantId}/entities/{entityId}/checks/{checkId}": {
                "href-template": "/v1.0/{tenantId}/entities/{entityId}/checks/{checkId}",
                "href-vars": {
                    "tenantId": "http://docs.rackspace.com/cm/api/v1.0/cm-devguide/params/tenantId",
                    "entityId": "http://docs.rackspace.com/cm/api/v1.0/cm-devguide/params/entityId",
                    "checkId": "http://docs.rackspace.com/cm/api/v1.0/cm-devguide/params/checkId"
                },
                "hints": {
                    "docs": ("http://docs.rackspace.com/cm/api/v1.0/cm-devguide"
                             "/content/service-checks.html"),
                    "allow": [
                        "GET",
                        "PUT",
                        "DELETE"
                    ],
                    "representations": [
                        "application/json",
                        "application/xml"
                    ],
                    "accept-put": [
                        "application/xml",
                        "application/json"
                    ]
                }
            },
            url + "/{tenantId}/entities/{entityId}/checks/{checkId}/test": {
                "href-template": "/v1.0/{tenantId}/entities/{entityId}/checks/{checkId}/test",
                "href-vars": {
                    "tenantId": "http://docs.rackspace.com/cm/api/v1.0/cm-devguide/params/tenantId",
                    "entityId": "http://docs.rackspace.com/cm/api/v1.0/cm-devguide/params/entityId",
                    "checkId": "http://docs.rackspace.com/cm/api/v1.0/cm-devguide/params/checkId"
                },
                "hints": {
                    "docs": ("http://docs.rackspace.com/cm/api/v1.0/cm-devguide"
                             "/content/service-checks.html"),
                    "allow": [
                        "POST"
                    ],
                    "representations": [
                        "application/json",
                        "application/xml"
                    ],
                    "accept-post": [
                        "application/xml",
                        "application/json"
                    ]
                }
            },
            url + "/{tenantId}/entities/{entityId}/test-check": {
                "href-template": "/v1.0/{tenantId}/entities/{entityId}/test-check",
                "href-vars": {
                    "tenantId": "http://docs.rackspace.com/cm/api/v1.0/cm-devguide/params/tenantId",
                    "entityId": "http://docs.rackspace.com/cm/api/v1.0/cm-devguide/params/entityId"
                },
                "hints": {
                    "docs": ("http://docs.rackspace.com/cm/api/v1.0/cm-devguide"
                             "/content/service-checks.html"),
                    "allow": [
                        "POST"
                    ],
                    "representations": [
                        "application/json",
                        "application/xml"
                    ],
                    "accept-post": [
                        "application/xml",
                        "application/json"
                    ]
                }
            },
            url + "/{tenantId}/entities/{entityId}/alarms": {
                "href-template": "/v1.0/{tenantId}/entities/{entityId}/alarms",
                "href-vars": {
                    "tenantId": "http://docs.rackspace.com/cm/api/v1.0/cm-devguide/params/tenantId",
                    "entityId": "http://docs.rackspace.com/cm/api/v1.0/cm-devguide/params/entityId"
                },
                "hints": {
                    "docs": ("http://docs.rackspace.com/cm/api/v1.0/cm-devguide"
                             "/content/service-alarms.html"),
                    "allow": [
                        "GET",
                        "POST"
                    ],
                    "representations": [
                        "application/json",
                        "application/xml"
                    ],
                    "accept-post": [
                        "application/xml",
                        "application/json"
                    ]
                }
            },
            url + "/{tenantId}/entities/{entityId}/alarms/{alarmId}": {
                "href-template": "/v1.0/{tenantId}/entities/{entityId}/alarms/{alarmId}",
                "href-vars": {
                    "tenantId": "http://docs.rackspace.com/cm/api/v1.0/cm-devguide/params/tenantId",
                    "entityId": "http://docs.rackspace.com/cm/api/v1.0/cm-devguide/params/entityId",
                    "alarmId": "http://docs.rackspace.com/cm/api/v1.0/cm-devguide/params/alarmId"
                },
                "hints": {
                    "docs": ("http://docs.rackspace.com/cm/api/v1.0/cm-devguide"
                             "/content/service-alarms.html"),
                    "allow": [
                        "GET",
                        "PUT",
                        "DELETE"
                    ],
                    "representations": [
                        "application/json",
                        "application/xml"
                    ],
                    "accept-put": [
                        "application/xml",
                        "application/json"
                    ]
                }
            },
            url + "/{tenantId}/entities/{entityId}/test-alarm": {
                "href-template": "/v1.0/{tenantId}/entities/{entityId}/test-alarm",
                "href-vars": {
                    "tenantId": "http://docs.rackspace.com/cm/api/v1.0/cm-devguide/params/tenantId",
                    "entityId": "http://docs.rackspace.com/cm/api/v1.0/cm-devguide/params/entityId"
                },
                "hints": {
                    "docs": ("http://docs.rackspace.com/cm/api/v1.0/cm-devguide"
                             "/content/service-alarms.html"),
                    "allow": [
                        "POST"
                    ],
                    "representations": [
                        "application/json",
                        "application/xml"
                    ],
                    "accept-post": [
                        "application/xml",
                        "application/json"
                    ]
                }
            },
            url + "/{tenantId}/entities/{entityId}/agent/host_info/{handlerType}": {
                "href-template": "/v1.0/{tenantId}/entities/{entityId}/agent/host_info/{handlerType}",
                "href-vars": {
                    "tenantId": "http://docs.rackspace.com/cm/api/v1.0/cm-devguide/params/tenantId",
                    "entityId": "http://docs.rackspace.com/cm/api/v1.0/cm-devguide/params/entityId",
                    "handlerType": "http://docs.rackspace.com/cm/api/v1.0/cm-devguide/params/handlerType"
                },
                "hints": {
                    "docs": ("http://docs.rackspace.com/cm/api/v1.0/cm-devguide"
                             "/content/service-agents.html"),
                    "allow": [
                        "GET"
                    ],
                    "representations": [
                        "application/json",
                        "application/xml"
                    ]
                }
            },
            url + "/{tenantId}/entities/{entityId}/agent/check_types/{checkType}/targets": {
                "href-template": ("/v1.0/{tenantId}/entities/{entityId}/agent/"
                                  "check_types/{checkType}/targets"),
                "href-vars": {
                    "tenantId": "http://docs.rackspace.com/cm/api/v1.0/cm-devguide/params/tenantId",
                    "entityId": "http://docs.rackspace.com/cm/api/v1.0/cm-devguide/params/entityId",
                    "checkType": "http://docs.rackspace.com/cm/api/v1.0/cm-devguide/params/checkType"
                },
                "hints": {
                    "docs": ("http://docs.rackspace.com/cm/api/v1.0/cm-devguide"
                             "/content/service-agents.html"),
                    "allow": [
                        "GET"
                    ],
                    "representations": [
                        "application/json",
                        "application/xml"
                    ]
                }
            },
            url + "/{tenantId}/entities/{entityId}/alarms/{alarmId}/notification_history": {
                "href-template": ("/v1.0/{tenantId}/entities/{entityId}/alarms/"
                                  "{alarmId}/notification_history"),
                "href-vars": {
                    "tenantId": "http://docs.rackspace.com/cm/api/v1.0/cm-devguide/params/tenantId",
                    "entityId": "http://docs.rackspace.com/cm/api/v1.0/cm-devguide/params/entityId",
                    "alarmId": "http://docs.rackspace.com/cm/api/v1.0/cm-devguide/params/alarmId"
                },
                "hints": {
                    "docs": ("http://docs.rackspace.com/cm/api/v1.0/cm-devguide/"
                             "content/service-alarms.html"),
                    "allow": [
                        "GET"
                    ],
                    "representations": [
                        "application/json",
                        "application/xml"
                    ]
                }
            },
            url + "/{tenantId}/entities/{entityId}/alarms/{alarmId}/notification_history/{checkId}": {
                "href-template": ("/v1.0/{tenantId}/entities/{entityId}/alarms/"
                                  "{alarmId}/notification_history/{checkId}"),
                "href-vars": {
                    "tenantId": "http://docs.rackspace.com/cm/api/v1.0/cm-devguide/params/tenantId",
                    "entityId": "http://docs.rackspace.com/cm/api/v1.0/cm-devguide/params/entityId",
                    "alarmId": "http://docs.rackspace.com/cm/api/v1.0/cm-devguide/params/alarmId",
                    "checkId": "http://docs.rackspace.com/cm/api/v1.0/cm-devguide/params/checkId"
                },
                "hints": {
                    "docs": ("http://docs.rackspace.com/cm/api/v1.0/cm-devguide"
                             "/content/service-alarms.html"),
                    "allow": [
                        "GET"
                    ],
                    "representations": [
                        "application/json",
                        "application/xml"
                    ]
                }
            },
            url + ("/{tenantId}/entities/{entityId}/alarms/{alarmId}"
                   "/notification_history/{checkId}/{uuid}"): {
                "href-template": ("/v1.0/{tenantId}/entities/{entityId}/alarms/"
                                  "{alarmId}/notification_history/{checkId}/{uuid}"),
                "href-vars": {
                    "tenantId": "http://docs.rackspace.com/cm/api/v1.0/cm-devguide/params/tenantId",
                    "entityId": "http://docs.rackspace.com/cm/api/v1.0/cm-devguide/params/entityId",
                    "alarmId": "http://docs.rackspace.com/cm/api/v1.0/cm-devguide/params/alarmId",
                    "checkId": "http://docs.rackspace.com/cm/api/v1.0/cm-devguide/params/checkId",
                    "uuid": "http://docs.rackspace.com/cm/api/v1.0/cm-devguide/params/uuid"
                },
                "hints": {
                    "docs": ("http://docs.rackspace.com/cm/api/v1.0/cm-devguide"
                             "/content/service-alarms.html"),
                    "allow": [
                        "GET"
                    ],
                    "representations": [
                        "application/json",
                        "application/xml"
                    ]
                }
            },
            url + "/{tenantId}/check_types": {
                "href-template": "/v1.0/{tenantId}/check_types",
                "href-vars": {
                    "tenantId": "http://docs.rackspace.com/cm/api/v1.0/cm-devguide/params/tenantId"
                },
                "hints": {
                    "docs": ("http://docs.rackspace.com/cm/api/v1.0/cm-devguide"
                             "/content/service-check-types.html"),
                    "allow": [
                        "GET"
                    ],
                    "representations": [
                        "application/json",
                        "application/xml"
                    ]
                }
            },
            url + "/{tenantId}/check_types/{checkTypeId}": {
                "href-template": "/v1.0/{tenantId}/check_types/{checkTypeId}",
                "href-vars": {
                    "tenantId": "http://docs.rackspace.com/cm/api/v1.0/cm-devguide/params/tenantId",
                    "checkTypeId": "http://docs.rackspace.com/cm/api/v1.0/cm-devguide/params/checkTypeId"
                },
                "hints": {
                    "docs": ("http://docs.rackspace.com/cm/api/v1.0/cm-devguide"
                             "/content/service-check-types.html"),
                    "allow": [
                        "GET"
                    ],
                    "representations": [
                        "application/json",
                        "application/xml"
                    ]
                }
            },
            url + "/{tenantId}/notification_types": {
                "href-template": "/v1.0/{tenantId}/notification_types",
                "href-vars": {
                    "tenantId": "http://docs.rackspace.com/cm/api/v1.0/cm-devguide/params/tenantId"
                },
                "hints": {
                    "docs": ("http://docs.rackspace.com/cm/api/v1.0/cm-devguide"
                             "/content/service-notification-types.html"),
                    "allow": [
                        "GET"
                    ],
                    "representations": [
                        "application/json",
                        "application/xml"
                    ]
                }
            },
            url + "/{tenantId}/notification_types/{notificationTypeId}": {
                "href-template": "/v1.0/{tenantId}/notification_types/{notificationTypeId}",
                "href-vars": {
                    "tenantId": "http://docs.rackspace.com/cm/api/v1.0/cm-devguide/params/tenantId",
                    "notificationTypeId": ("http://docs.rackspace.com/cm/api/v1.0/cm-devguide"
                                           "/params/notificationTypeId")
                },
                "hints": {
                    "docs": ("http://docs.rackspace.com/cm/api/v1.0/cm-devguide"
                             "/content/service-notification-types.html"),
                    "allow": [
                        "GET"
                    ],
                    "representations": [
                        "application/json",
                        "application/xml"
                    ]
                }
            },
            url + "/{tenantId}/notifications": {
                "href-template": "/v1.0/{tenantId}/notifications",
                "href-vars": {
                    "tenantId": "http://docs.rackspace.com/cm/api/v1.0/cm-devguide/params/tenantId"
                },
                "hints": {
                    "docs": ("http://docs.rackspace.com/cm/api/v1.0/cm-devguide"
                             "/content/service-notifications.html"),
                    "allow": [
                        "GET",
                        "POST"
                    ],
                    "representations": [
                        "application/json",
                        "application/xml"
                    ],
                    "accept-post": [
                        "application/xml",
                        "application/json"
                    ]
                }
            },
            url + "/{tenantId}/notifications/{notificationId}": {
                "href-template": "/v1.0/{tenantId}/notifications/{notificationId}",
                "href-vars": {
                    "tenantId": "http://docs.rackspace.com/cm/api/v1.0/cm-devguide/params/tenantId",
                    "notificationId": ("http://docs.rackspace.com/cm/api/v1.0/cm-devguide"
                                       "/params/notificationId")
                },
                "hints": {
                    "docs": ("http://docs.rackspace.com/cm/api/v1.0/cm-devguide"
                             "/content/service-notifications.html"),
                    "allow": [
                        "GET",
                        "PUT",
                        "DELETE"
                    ],
                    "representations": [
                        "application/json",
                        "application/xml"
                    ],
                    "accept-put": [
                        "application/xml",
                        "application/json"
                    ]
                }
            },
            url + "/{tenantId}/notifications/{notificationId}/test": {
                "href-template": "/v1.0/{tenantId}/notifications/{notificationId}/test",
                "href-vars": {
                    "tenantId": "http://docs.rackspace.com/cm/api/v1.0/cm-devguide/params/tenantId",
                    "notificationId": ("http://docs.rackspace.com/cm/api/v1.0/cm-devguide"
                                       "/params/notificationId")
                },
                "hints": {
                    "docs": ("http://docs.rackspace.com/cm/api/v1.0/cm-devguide"
                             "/content/service-notifications.html"),
                    "allow": [
                        "POST"
                    ],
                    "representations": [
                        "application/json",
                        "application/xml"
                    ],
                    "accept-post": [
                        "application/xml",
                        "application/json"
                    ]
                }
            },
            url + "/{tenantId}/test-notification": {
                "href-template": "/v1.0/{tenantId}/test-notification",
                "href-vars": {
                    "tenantId": "http://docs.rackspace.com/cm/api/v1.0/cm-devguide/params/tenantId"
                },
                "hints": {
                    "docs": ("http://docs.rackspace.com/cm/api/v1.0/cm-devguide"
                             "/content/service-notifications.html"),
                    "allow": [
                        "POST"
                    ],
                    "representations": [
                        "application/json",
                        "application/xml"
                    ],
                    "accept-post": [
                        "application/xml",
                        "application/json"
                    ]
                }
            },
            url + "/{tenantId}/notification_plans": {
                "href-template": "/v1.0/{tenantId}/notification_plans",
                "href-vars": {
                    "tenantId": "http://docs.rackspace.com/cm/api/v1.0/cm-devguide/params/tenantId"
                },
                "hints": {
                    "docs": ("http://docs.rackspace.com/cm/api/v1.0/cm-devguide"
                             "/content/service-notification-plans.html"),
                    "allow": [
                        "GET",
                        "POST"
                    ],
                    "representations": [
                        "application/json",
                        "application/xml"
                    ],
                    "accept-post": [
                        "application/xml",
                        "application/json"
                    ]
                }
            },
            url + "/{tenantId}/notification_plans/{notificationPlanId}": {
                "href-template": "/v1.0/{tenantId}/notification_plans/{notificationPlanId}",
                "href-vars": {
                    "tenantId": "http://docs.rackspace.com/cm/api/v1.0/cm-devguide/params/tenantId",
                    "notificationPlanId": ("http://docs.rackspace.com/cm/api/v1.0/cm-devguide"
                                           "/params/notificationPlanId")
                },
                "hints": {
                    "docs": ("http://docs.rackspace.com/cm/api/v1.0/cm-devguide"
                             "/content/service-notification-plans.html"),
                    "allow": [
                        "GET",
                        "PUT",
                        "DELETE"
                    ],
                    "representations": [
                        "application/json",
                        "application/xml"
                    ],
                    "accept-put": [
                        "application/xml",
                        "application/json"
                    ]
                }
            },
            url + "/{tenantId}/audits": {
                "href-template": "/v1.0/{tenantId}/audits",
                "href-vars": {
                    "tenantId": "http://docs.rackspace.com/cm/api/v1.0/cm-devguide/params/tenantId"
                },
                "hints": {
                    "docs": ("http://docs.rackspace.com/cm/api/v1.0/cm-devguide"
                             "/content/service-audits.html"),
                    "allow": [
                        "GET"
                    ],
                    "representations": [
                        "application/json",
                        "application/xml"
                    ]
                }
            },
            url + "/{tenantId}/limits": {
                "href-template": "/v1.0/{tenantId}/limits",
                "href-vars": {
                    "tenantId": "http://docs.rackspace.com/cm/api/v1.0/cm-devguide/params/tenantId"
                },
                "hints": {
                    "docs": ("http://docs.rackspace.com/cm/api/v1.0/cm-devguide"
                             "/content/service-limits.html"),
                    "allow": [
                        "GET"
                    ],
                    "representations": [
                        "application/json",
                        "application/xml"
                    ]
                }
            },
            url + "/{tenantId}/usage": {
                "href-template": "/v1.0/{tenantId}/usage",
                "href-vars": {
                    "tenantId": "http://docs.rackspace.com/cm/api/v1.0/cm-devguide/params/tenantId"
                },
                "hints": {
                    "docs": ("http://docs.rackspace.com/cm/api/v1.0/cm-devguide"
                             "/content/service-usage.html"),
                    "allow": [
                        "GET"
                    ],
                    "representations": [
                        "application/json",
                        "application/xml"
                    ]
                }
            },
            url + "/{tenantId}/suppressions/{suppressionId}": {
                "href-template": "/v1.0/{tenantId}/suppressions/{suppressionId}",
                "href-vars": {
                    "tenantId": "http://docs.rackspace.com/cm/api/v1.0/cm-devguide/params/tenantId",
                    "suppressionId": ("http://docs.rackspace.com/cm/api/v1.0/cm-devguide"
                                      "/params/suppressionId")
                },
                "hints": {
                    "docs": ("http://docs.rackspace.com/cm/api/v1.0/cm-devguide"
                             "/content/service-suppressions.html"),
                    "allow": [
                        "GET",
                        "PUT",
                        "DELETE"
                    ],
                    "representations": [
                        "application/json",
                        "application/xml"
                    ],
                    "accept-put": [
                        "application/xml",
                        "application/json"
                    ]
                }
            },
            url + "/{tenantId}/suppressions": {
                "href-template": "/v1.0/{tenantId}/suppressions",
                "href-vars": {
                    "tenantId": "http://docs.rackspace.com/cm/api/v1.0/cm-devguide/params/tenantId"
                },
                "hints": {
                    "docs": ("http://docs.rackspace.com/cm/api/v1.0/cm-devguide"
                             "/content/service-suppressions.html"),
                    "allow": [
                        "POST",
                        "GET"
                    ],
                    "representations": [
                        "application/json",
                        "application/xml"
                    ],
                    "accept-post": [
                        "application/xml",
                        "application/json"
                    ]
                }
            },
            url + "/{tenantId}/suppression_logs": {
                "href-template": "/v1.0/{tenantId}/suppression_logs",
                "href-vars": {
                    "tenantId": "http://docs.rackspace.com/cm/api/v1.0/cm-devguide/params/tenantId"
                },
                "hints": {
                    "docs": ("http://docs.rackspace.com/cm/api/v1.0/cm-devguide"
                             "/content/service-suppression-log_entries.html"),
                    "allow": [
                        "GET"
                    ],
                    "representations": [
                        "application/json",
                        "application/xml"
                    ]
                }
            },
            url + "/{tenantId}/alarm_examples": {
                "href-template": "/v1.0/{tenantId}/alarm_examples",
                "href-vars": {
                    "tenantId": "http://docs.rackspace.com/cm/api/v1.0/cm-devguide/params/tenantId"
                },
                "hints": {
                    "docs": ("http://docs.rackspace.com/cm/api/v1.0/cm-devguide"
                             "/content/service-alarm-examples.html"),
                    "allow": [
                        "GET"
                    ],
                    "representations": [
                        "application/json",
                        "application/xml"
                    ]
                }
            },
            url + "/{tenantId}/alarm_examples/{alarmExampleId}": {
                "href-template": "/v1.0/{tenantId}/alarm_examples/{alarmExampleId}",
                "href-vars": {
                    "tenantId": "http://docs.rackspace.com/cm/api/v1.0/cm-devguide/params/tenantId",
                    "alarmExampleId": ("http://docs.rackspace.com/cm/api/v1.0/cm-devguide"
                                       "/params/alarmExampleId")
                },
                "hints": {
                    "docs": ("http://docs.rackspace.com/cm/api/v1.0/cm-devguide"
                             "/content/service-alarm-examples.html"),
                    "allow": [
                        "GET",
                        "POST"
                    ],
                    "representations": [
                        "application/json",
                        "application/xml"
                    ],
                    "accept-post": [
                        "application/xml",
                        "application/json"
                    ]
                }
            },
            url + "/{tenantId}/views/overview": {
                "href-template": "/v1.0/{tenantId}/views/overview",
                "href-vars": {
                    "tenantId": "http://docs.rackspace.com/cm/api/v1.0/cm-devguide/params/tenantId"
                },
                "hints": {
                    "docs": ("http://docs.rackspace.com/cm/api/v1.0/cm-devguide"
                             "/content/service-views.html"),
                    "allow": [
                        "GET"
                    ],
                    "representations": [
                        "application/json",
                        "application/xml"
                    ]
                }
            },
            url + "/{tenantId}/views/metric_list": {
                "href-template": "/v1.0/{tenantId}/views/metric_list",
                "href-vars": {
                    "tenantId": "http://docs.rackspace.com/cm/api/v1.0/cm-devguide/params/tenantId"
                },
                "hints": {
                    "docs": ("http://docs.rackspace.com/cm/api/v1.0/cm-devguide"
                             "/content/service-views.html"),
                    "allow": [
                        "GET"
                    ],
                    "representations": [
                        "application/json",
                        "application/xml"
                    ]
                }
            },
            url + "/{tenantId}/views/agent_host_info": {
                "href-template": "/v1.0/{tenantId}/views/agent_host_info",
                "href-vars": {
                    "tenantId": "http://docs.rackspace.com/cm/api/v1.0/cm-devguide/params/tenantId"
                },
                "hints": {
                    "docs": ("http://docs.rackspace.com/cm/api/v1.0/cm-devguide"
                             "/content/service-views.html"),
                    "allow": [
                        "GET"
                    ],
                    "representations": [
                        "application/json",
                        "application/xml"
                    ]
                }
            },
            url + "/{tenantId}/views/latest_alarm_states": {
                "href-template": "/v1.0/{tenantId}/views/latest_alarm_states",
                "href-vars": {
                    "tenantId": "http://docs.rackspace.com/cm/api/v1.0/cm-devguide/params/tenantId"
                },
                "hints": {
                    "docs": ("http://docs.rackspace.com/cm/api/v1.0/cm-devguide"
                             "/content/service-views.html"),
                    "allow": [
                        "GET"
                    ],
                    "representations": [
                        "application/json",
                        "application/xml"
                    ]
                }
            },
            url + "/{tenantId}/__experiments/json_home": {
                "href-template": "/v1.0/{tenantId}/__experiments/json_home",
                "href-vars": {
                    "tenantId": "http://docs.rackspace.com/cm/api/v1.0/cm-devguide/params/tenantId"
                },
                "hints": {
                    "docs": ("http://docs.rackspace.com/cm/api/v1.0/cm-devguide"
                             "/content/service-json-home.html"),
                    "allow": [
                        "GET"
                    ],
                    "representations": [
                        "application/json",
                        "application/xml"
                    ]
                }
            },
            url + "/{tenantId}/__experiments/multiplot": {
                "href-template": "/v1.0/{tenantId}/__experiments/multiplot",
                "href-vars": {
                    "tenantId": "http://docs.rackspace.com/cm/api/v1.0/cm-devguide/params/tenantId"
                },
                "hints": {
                    "docs": ("http://docs.rackspace.com/cm/api/v1.0/cm-devguide"
                             "/content/service-rollups.html"),
                    "allow": [
                        "POST"
                    ],
                    "representations": [
                        "application/json",
                        "application/xml"
                    ],
                    "accept-post": [
                        "application/xml",
                        "application/json"
                    ]
                }
            },
            url + "/{tenantId}/entities/{entityId}/checks/{checkId}/metrics/{metricName}/plot": {
                "href-template": ("/v1.0/{tenantId}/entities/{entityId}/checks"
                                  "/{checkId}/metrics/{metricName}/plot"),
                "href-vars": {
                    "tenantId": "http://docs.rackspace.com/cm/api/v1.0/cm-devguide/params/tenantId",
                    "entityId": "http://docs.rackspace.com/cm/api/v1.0/cm-devguide/params/entityId",
                    "checkId": "http://docs.rackspace.com/cm/api/v1.0/cm-devguide/params/checkId",
                    "metricName": "http://docs.rackspace.com/cm/api/v1.0/cm-devguide/params/metricName"
                },
                "hints": {
                    "docs": ("http://docs.rackspace.com/cm/api/v1.0/cm-devguide"
                             "/content/service-rollups.html"),
                    "allow": [
                        "GET"
                    ],
                    "representations": [
                        "application/json",
                        "application/xml"
                    ]
                }
            },
            url + "/{tenantId}/entities/{entityId}/checks/{checkId}/metrics": {
                "href-template": "/v1.0/{tenantId}/entities/{entityId}/checks/{checkId}/metrics",
                "href-vars": {
                    "tenantId": "http://docs.rackspace.com/cm/api/v1.0/cm-devguide/params/tenantId",
                    "entityId": "http://docs.rackspace.com/cm/api/v1.0/cm-devguide/params/entityId",
                    "checkId": "http://docs.rackspace.com/cm/api/v1.0/cm-devguide/params/checkId"
                },
                "hints": {
                    "docs": ("http://docs.rackspace.com/cm/api/v1.0/cm-devguide"
                             "/content/service-metric-names.html"),
                    "allow": [
                        "GET"
                    ],
                    "representations": [
                        "application/json",
                        "application/xml"
                    ]
                }
            }
        }
    }



"""
Setup file for mimic
"""
from __future__ import print_function

from setuptools import setup, find_packages

try:
    from py2app.build_app import py2app
    py2app_available = True
except ImportError:
    py2app_available = False


_NAME = "mimic"
_VERSION = "1.9.0"


def setup_options(name, version):
    """
    If `py2app` is present in path, then enable to option to build the app.

    This also disables the options needed for normal `sdist` installs.

    :returns: a dictionary of setup options.
    """
    info = dict(
        install_requires=[
            "characteristic>=14.2.0",
            "klein>=0.2.1",
            "twisted>=14.0.2",
            "jsonschema>=2.0",
            "treq>=0.2.1",
            "six>=1.6.0",
            "xmltodict>=0.9.1",
            "attrs>=15.0.0",
            "testtools>=1.7.1,<1.8.0",
            "iso8601>=0.1.10",
            "toolz>=0.7.4"
        ],
        package_dir={"mimic": "mimic"},
        packages=find_packages(exclude=[]) + ["twisted.plugins"],
    )
    if not py2app_available:
        return info

    # py2app available, proceed.
    script="bundle/start-app.py"
    test_script="bundle/run-tests.py"
    plist = dict(
        CFBundleName                = _NAME,
        CFBundleShortVersionString  = " ".join([_NAME, _VERSION]),
        CFBundleGetInfoString       = _NAME,
        CFBundleExecutable          = _NAME,
        CFBundleIdentifier          = "com.%s.%s" % (_NAME, _VERSION),
        LSUIElement                 = "1",
        LSMultipleInstancesProhibited = "1",
    )
    app_data = dict(
        script=script,
        plist=plist,
        extra_scripts=[test_script]
    )

    class BuildWithCache(py2app, object):
        """
        Before building the application rebuild the `dropin.cache` files.
        """

        def collect_recipedict(self):
            """
            Implement a special Twisted plugins recipe so that dropin.cache
            files are generated and included in site-packages.zip.
            """
            result = super(BuildWithCache, self).collect_recipedict()
            def check(cmd, mg):
                from twisted.plugin import getPlugins, IPlugin
                from twisted import plugins as twisted_plugins
                from mimic import plugins as mimic_plugins

                for plugin_package in [twisted_plugins, mimic_plugins]:
                    import time
                    list(getPlugins(IPlugin, package=plugin_package))

                import os
                def plugpath(what):
                    path_in_zip = what + "/plugins"
                    path_on_fs = (
                        os.path.abspath(
                            os.path.join(
                                os.path.dirname(
                                    __import__(what + ".plugins",
                                               fromlist=["nonempty"])
                                    .__file__),
                                "dropin.cache")
                        ))
                    os.utime(path_on_fs, (time.time() + 86400,) * 2)
                    return (path_in_zip, [path_on_fs])
                data_files = [plugpath("mimic"), plugpath("twisted")]

                return dict(loader_files=data_files)
            result["bonus"] = check
            return result

    return dict(
        info,
        app=[app_data],
        cmdclass={
            "py2app": BuildWithCache
        },
        options={
            "py2app": {
                "includes": [
                    "syslog",
                    "mimic.test.*",
                    "mimic.plugins.*",
                    "twisted.plugins.*",
                    "twisted.plugin",
                ],
            }
        }
    )

setup(
    name=_NAME,
    version=_VERSION,
    description="An API-compatible mock service",
    license="Apache License, Version 2.0",
    url="https://github.com/rackerlabs/mimic",
    include_package_data=True,
    **setup_options(_NAME, _VERSION)
)



"""
MAAS Mock API
"""

from __future__ import division

import json
import collections
import urlparse
import random
import re
from uuid import uuid4

from six import text_type

from characteristic import attributes
from zope.interface import implementer

from twisted.plugin import IPlugin

from mimic.catalog import Entry
from mimic.catalog import Endpoint
from mimic.rest.auth_api import base_uri_from_request
from mimic.rest.mimicapp import MimicApp
from mimic.imimic import IAPIMock
from mimic.canned_responses.maas_json_home import json_home
from mimic.canned_responses.maas_agent_info import agent_info
from mimic.canned_responses.maas_monitoring_zones import monitoring_zones
from mimic.canned_responses.maas_alarm_examples import alarm_examples
from mimic.model.maas_objects import (Alarm,
                                      AlarmState,
                                      Check,
                                      Entity,
                                      MaasStore,
                                      Notification,
                                      NotificationPlan,
                                      Suppression)
from mimic.util.helper import Matcher, random_hex_generator, random_hipsum


MISSING_REQUIRED_KEY_REGEX = re.compile(r'Missing keyword value for \'(\w+)\'.')
REMOTE_CHECK_TYPE_REGEX = re.compile(r'^remote\.')


@implementer(IAPIMock, IPlugin)
class MaasApi(object):

    """
    Rest endpoints for mocked MAAS Api.
    """

    def __init__(self, regions=["ORD"]):
        """
        Set regions
        """
        self._regions = regions

    def catalog_entries(self, tenant_id):
        """
        List catalog entries for the MaaS API.
        """
        return [
            Entry(
                tenant_id, "rax: monitor", "cloudMonitoring",
                [
                    Endpoint(tenant_id, region, text_type(uuid4()),
                             "v1.0")
                    for region in self._regions
                ]
            )
        ]

    def resource_for_region(self, region, uri_prefix, session_store):
        """
        Get an : obj: `twisted.web.iweb.IResource` for the given URI prefix;
        implement : obj: `IAPIMock`.
        """
        return MaasMock(self, uri_prefix, session_store, region).app.resource()


class MCache(object):

    """
    M(onitoring) Cache Object to hold dictionaries of all entities, checks and alarms.
    """

    def __init__(self, clock):
        """
        Create the initial structs for cache
        """
        current_time_milliseconds = int(1000 * clock.seconds())

        self.agenthostinfo_querycount = collections.defaultdict(lambda: 0, {})
        self.entities_list = []
        self.checks_list = []
        self.alarms_list = []
        self.notifications_list = [Notification(id=u'ntTechnicalContactsEmail',
                                                label=u'Email All Technical Contacts',
                                                created_at=current_time_milliseconds,
                                                updated_at=current_time_milliseconds,
                                                type=u'technicalContactsEmail')]
        self.notificationplans_list = [NotificationPlan(id=u'npTechnicalContactsEmail',
                                                        label=u'Technical Contacts - Email',
                                                        created_at=current_time_milliseconds,
                                                        updated_at=current_time_milliseconds)]
        self.notificationtypes_list = [{'id': 'webhook', 'fields': [{'name': 'url',
                                                                     'optional': False,
                                                                     'description': 'An HTTP or \
                                                                      HTTPS URL to POST to'}]},
                                       {'id': 'email', 'fields': [{'name': 'address',
                                                                   'optional': False,
                                                                   'description': 'Email \
                                                                    address to send notifications to'}]},
                                       {'id': 'pagerduty', 'fields': [{'name': 'service_key',
                                                                       'optional': False,
                                                                       'description': 'The PagerDuty \
                                                                        service key to use.'}]},
                                       {'id': 'sms', 'fields': [{'name': 'phone_number',
                                                                 'optional': False,
                                                                 'description': 'Phone number to send \
                                                                  the notification to, \
                                                                  with leading + and country \
                                                                  code (E.164 format)'}]}]
        self.suppressions_list = []
        self.audits_list = []
        self.maas_store = MaasStore(clock)
        self.test_alarm_responses = {}
        self.test_alarm_errors = {}


def _only_keys(dict_ins, keys):
    """
    Filters out unwanted keys of a dict.
    """
    return dict([(k, dict_ins[k]) for k in dict_ins if k in keys])


def create_entity(clock, params):
    """
    Returns a dictionary representing an entity

    :return: an Entity model, which is described in `the Rackspace Cloud
        Monitoring Developer Guide, section 5.4
        <http://docs.rackspace.com/cm/api/v1.0/cm-devguide/content/service-entities.html>`_
    :rtype: ``dict`` mapping ``unicode`` to ``unicode``, ``float``,
        ``bool``, ``dict`` or ``NoneType``.
    """
    current_time_milliseconds = int(1000 * clock.seconds())
    params_copy = _only_keys(params, Entity.USER_SPECIFIABLE_KEYS)
    params_copy['created_at'] = params_copy['updated_at'] = current_time_milliseconds
    return Entity(**params_copy)


def create_check(clock, entity_id, params):
    """
    Returns a dictionary representing a check

    :return: a Check model, which is described in `the Rackspace Cloud
        Monitoring Developer Guide, section 5.7
        <http://docs.rackspace.com/cm/api/v1.0/cm-devguide/content/service-checks.html>`_
    :rtype: ``dict`` mapping ``unicode`` to ``unicode``, ``float``,
        ``int``, ``bool``, ``dict`` or ``NoneType``.
    """
    current_time_milliseconds = int(1000 * clock.seconds())
    params_copy = _only_keys(params, Check.USER_SPECIFIABLE_KEYS)
    params_copy['entity_id'] = entity_id
    params_copy['created_at'] = params_copy['updated_at'] = current_time_milliseconds
    return Check(**params_copy)


def create_alarm(clock, entity_id, params):
    """
    Returns a dictionary representing an alarm

    :return: an Alarm model, which is described in `the Rackspace Cloud
        Monitoring Developer Guide, section 5.12
        <http://docs.rackspace.com/cm/api/v1.0/cm-devguide/content/service-alarms.html>`_
    :rtype: ``dict`` mapping ``unicode`` to ``unicode``, ``float``,
        ``bool``, ``dict``, or ``NoneType``.
    """
    current_time_milliseconds = int(1000 * clock.seconds())
    params_copy = _only_keys(params, Alarm.USER_SPECIFIABLE_KEYS)
    params_copy['entity_id'] = entity_id
    params_copy['created_at'] = params_copy['updated_at'] = current_time_milliseconds
    return Alarm(**params_copy)


def create_notification_plan(clock, params):
    """
    Creates a notification plan

    :return: a Notification Plan model, which is described in `the
        Rackspace Cloud Monitoring Developer Guide, section 5.11
        <http://docs.rackspace.com/cm/api/v1.0/cm-devguide/content/service-notification-plans.html>`_
    :rtype: ``dict`` mapping ``unicode`` to ``unicode``, ``float``,
        ``dict`` or ``NoneType``.
    """
    current_time_milliseconds = int(1000 * clock.seconds())
    params_copy = _only_keys(params, NotificationPlan.USER_SPECIFIABLE_KEYS)
    params_copy['created_at'] = params_copy['updated_at'] = current_time_milliseconds
    return NotificationPlan(**params_copy)


def create_notification(clock, params):
    """
    Creates a notification target

    :return: a Notification model, which is described in `the Rackspace
        Cloud Monitoring Developer Guide, section 5.10
        <http://docs.rackspace.com/cm/api/v1.0/cm-devguide/content/service-notifications.html>`_
    :rtype: ``dict`` mapping ``unicode`` to ``unicode``, ``float``,
        ``dict`` or ``NoneType``.
    """
    current_time_milliseconds = int(1000 * clock.seconds())
    params_copy = _only_keys(params, Notification.USER_SPECIFIABLE_KEYS)
    params_copy['created_at'] = params_copy['updated_at'] = current_time_milliseconds
    return Notification(**params_copy)


def create_suppression(clock, params):
    """
    Creates a suppression

    :return: a Suppression model, which is described in `the Rackspace
        Cloud Monitoring Developer Guide, section 5.16
        <http://docs.rackspace.com/cm/api/v1.0/cm-devguide/content/service-suppressions.html>`_
    :rtype: ``dict`` mapping ``unicode`` to ``unicode`` or ``list``.
    """
    params_copy = _only_keys(params, Suppression.USER_SPECIFIABLE_KEYS)
    params_copy['created_at'] = params_copy['updated_at'] = int(1000 * clock.seconds())
    return Suppression(**params_copy)


def _object_getter(collection, matcher, request, object_type, object_key):
    """
    Generic getter handler for objects in a collection.
    """
    for obj in collection:
        if matcher(obj):
            request.setResponseCode(200)
            return json.dumps(obj.to_json())
    request.setResponseCode(404)
    return json.dumps({'type': 'notFoundError',
                       'code': 404,
                       'txnId': '.fake.mimic.transaction.id.c-1111111.ts-123444444.v-12344frf',
                       'message': 'Object does not exist',
                       'details': 'Object "{0}" with key "{1}" does not exist'.format(
                           object_type, object_key)
                       })


def _metric_list_for_check(maas_store, entity, check):
    """
    Computes the metrics list for a given check.

    Remote checks return a metric for each monitoring zone and
    each type of metric for the check type. Agent checks return
    a metric for each metric type on the check type. Check types
    that Mimic doesn't know about generate an empty list.
    """
    if check.type not in maas_store.check_types:
        return []

    if REMOTE_CHECK_TYPE_REGEX.match(check.type):
        return [{'name': '{0}.{1}'.format(mz, metric.name),
                 'type': metric.type,
                 'unit': metric.unit}
                for metric in maas_store.check_types[check.type].metrics
                for mz in check.monitoring_zones_poll]

    return [{'name': metric.name,
             'type': metric.type,
             'unit': metric.unit}
            for metric in maas_store.check_types[check.type].metrics]


def _metric_list_for_entity(maas_store, entity, checks):
    """
    Creates the metrics list for one entity.
    """
    return {'entity_id': entity.id,
            'entity_label': entity.label,
            'checks': [{'id': check.id,
                        'label': check.label,
                        'type': check.type,
                        'metrics': _metric_list_for_check(maas_store, entity, check)}
                       for check in checks if check.entity_id == entity.id]}


def _multiplot_interval(from_date, to_date, points):
    """
    Computes the size of the interval between points in a multiplot.

    :return: the multiplot interval size.
    :rtype: ``float``
    """
    if points < 2:
        return 0.0
    return (to_date - from_date) / (points - 1)


def _compute_multiplot(maas_store, entity_id, check, metric_name, from_date, to_date, points):
    """
    Computes multiplot data for a single (entity, check, metric) group.
    """
    fallback = {'entity_id': entity_id,
                'check_id': check.id,
                'metric': metric_name,
                'unit': 'unknown',
                'type': 'unknown',
                'data': []}

    if check.type not in maas_store.check_types:
        return fallback

    interval = _multiplot_interval(from_date, to_date, points)
    metric = None
    base_metric_name = metric_name
    metric_value_kwargs = {'entity_id': entity_id,
                           'check_id': check.id}
    if re.match(r'^remote\.', check.type):
        match = re.match(r'^(mz\w+)\.(\w+)$', metric_name)
        if not match:
            return fallback
        metric_value_kwargs['monitoring_zone'] = match.group(1)
        base_metric_name = match.group(2)

    try:
        metric = maas_store.check_types[check.type].get_metric_by_name(base_metric_name)
    except NameError:
        return fallback

    return {'entity_id': entity_id,
            'check_id': check.id,
            'metric': metric_name,
            'unit': metric.unit,
            'type': metric.type,
            'data': [{'numPoints': 4,
                      'timestamp': int(from_date + (i * interval)),
                      'average': metric.get_value(
                          timestamp=int(from_date + (i * interval)),
                          **metric_value_kwargs)}
                     for i in range(points)]}


def parse_and_flatten_qs(url):
    """
    Parses a querystring and flattens 1-arg arrays.
    """
    qs = urlparse.parse_qs(url)
    flat_qs = {}
    for key in qs:
        flat_qs[key] = qs[key][0] if len(qs[key]) == 1 else qs[key]
    return flat_qs


def _mcache_factory(clock):
    """
    Returns a function that makes a defaultdict that makes MCache objects
    for each tenant.
    """
    return lambda: collections.defaultdict(lambda: MCache(clock))


class MaasMock(object):

    """
    Klein routes for the Monitoring API.
    """

    def __init__(self, api_mock, uri_prefix, session_store, name):
        """
        Create a maas region with a given URI prefix (used for generating URIs
        to servers).
        """
        self._api_mock = api_mock
        self._session_store = session_store
        self._name = name

    def _entity_cache_for_tenant(self, tenant_id):
        """
        Retrieve the M_cache object containing all objects created so far
        """
        clock = self._session_store.clock
        return (self._session_store.session_for_tenant_id(tenant_id)
                .data_for_api(self._api_mock, _mcache_factory(clock))[self._name]
                )

    def _audit(self, app, request, tenant_id, status, content=''):
        headers = dict([(k, v) for k, v in request.getAllHeaders().iteritems()
                        if k != 'x-auth-token'])

        self._entity_cache_for_tenant(tenant_id).audits_list.append(
            {
                'id': str(uuid4()),
                'timestamp': int(1000 * self._session_store.clock.seconds()),
                'headers': headers,
                'url': request.path,
                'app': app,
                'query': parse_and_flatten_qs(request.uri),
                'txnId': str(uuid4()),
                'payload': content,
                'method': request.method,
                'account_id': tenant_id,
                'who': '',
                'why': '',
                'statusCode': status
            })

    app = MimicApp()

    @app.route('/v1.0/<string:tenant_id>/mimic/reset', methods=['GET'])
    def doreset(self, request, tenant_id):
        """
        Reset the session
        """
        self._session_store.session_for_tenant_id(tenant_id)._api_objects = {}
        return "Session has been reset for tenant_id " + tenant_id

    @app.route('/v1.0/<string:tenant_id>/entities', methods=['GET'])
    def list_entities(self, request, tenant_id):
        """
        Replies the entities list call
        """
        entities = self._entity_cache_for_tenant(tenant_id).entities_list

        limit = 100
        marker = None
        next_marker = None
        next_href = None
        if 'limit' in request.args:
            limit = int(request.args['limit'][0].strip())
        if 'marker' in request.args:
            marker = request.args['marker'][0].strip()
            for q in range(len(entities)):
                if entities[q].id == marker:
                    entities = entities[q:]
                    break
        try:
            next_marker = entities[limit].id
        except Exception:
            pass
        entities = entities[:limit]

        metadata = {'count': len(entities),
                    'limit': limit,
                    'marker': marker,
                    'next_marker': next_marker,
                    'next_href': next_href}
        request.setResponseCode(200)
        return json.dumps({'metadata': metadata,
                           'values': [entity.to_json() for entity in entities]})

    @app.route('/v1.0/<string:tenant_id>/entities', methods=['POST'])
    def create_entity(self, request, tenant_id):
        """
        Creates a new entity
        """
        content = request.content.read()
        postdata = json.loads(content)
        newentity = create_entity(self._session_store.clock, postdata)
        self._entity_cache_for_tenant(tenant_id).entities_list.append(newentity)
        status = 201
        request.setResponseCode(status)
        request.setHeader('location', base_uri_from_request(request).rstrip('/') +
                          request.path + '/' + newentity.id.encode('utf-8'))
        request.setHeader('x-object-id', newentity.id.encode('utf-8'))
        request.setHeader('content-type', 'text/plain')
        self._audit('entities', request, tenant_id, status, content)
        return b''

    @app.route('/v1.0/<string:tenant_id>/entities/<string:entity_id>', methods=['GET'])
    def get_entity(self, request, tenant_id, entity_id):
        """
        Fetches a specific entity
        """
        return _object_getter(self._entity_cache_for_tenant(tenant_id).entities_list,
                              lambda entity: entity.id == entity_id,
                              request,
                              "Entity",
                              entity_id)

    @app.route('/v1.0/<string:tenant_id>/entities/<string:entity_id>/checks', methods=['GET'])
    def get_checks_for_entity(self, request, tenant_id, entity_id):
        """
        Returns all the checks for a paricular entity
        """
        checks = [check.to_json()
                  for check in self._entity_cache_for_tenant(tenant_id).checks_list
                  if check.entity_id == entity_id]
        metadata = {'count': len(checks),
                    'limit': 1000,
                    'marker': None,
                    'next_marker': None,
                    'next_href': None}
        request.setResponseCode(200)
        return json.dumps({'metadata': metadata, 'values': checks})

    @app.route('/v1.0/<string:tenant_id>/entities/<string:entity_id>', methods=['PUT'])
    def update_entity(self, request, tenant_id, entity_id):
        """
        Update entity in place.
        """
        content = request.content.read()
        update = json.loads(content)
        update_kwargs = dict(update)
        update_kwargs['clock'] = self._session_store.clock
        for entity in self._entity_cache_for_tenant(tenant_id).entities_list:
            if entity.id == entity_id:
                entity.update(**update_kwargs)
                break

        status = 204
        request.setResponseCode(status)
        request.setHeader('location', base_uri_from_request(request).rstrip('/') + request.path)
        request.setHeader('x-object-id', entity_id.encode('utf-8'))
        request.setHeader('content-type', 'text/plain')
        self._audit('entities', request, tenant_id, status, content)
        return b''

    @app.route('/v1.0/<string:tenant_id>/entities/<string:entity_id>', methods=['DELETE'])
    def delete_entity(self, request, tenant_id, entity_id):
        """
        Delete an entity, all checks that belong to entity, all alarms that belong to those checks
        """
        entities = self._entity_cache_for_tenant(tenant_id).entities_list

        try:
            entities.remove(Matcher(lambda entity: entity.id == entity_id))
        except ValueError:
            status = 404
            request.setResponseCode(status)
            self._audit('entities', request, tenant_id, status)
            return json.dumps({'type': 'notFoundError',
                               'code': status,
                               'message': 'Object does not exist',
                               'details': 'Object "Entity" with key "{0}" does not exist'.format(
                                   entity_id),
                               'txnId': '.fake.mimic.transaction.id.c-1111111.ts-123444444.v-12344frf'})

        checks = self._entity_cache_for_tenant(tenant_id).checks_list
        checks[:] = [check for check in checks
                     if check.entity_id != entity_id]

        alarms = self._entity_cache_for_tenant(tenant_id).alarms_list
        alarms[:] = [alarm for alarm in alarms
                     if alarm.entity_id != entity_id]

        status = 204
        request.setResponseCode(status)
        request.setHeader('content-type', 'text/plain')
        self._audit('entities', request, tenant_id, status)
        return b''

    @app.route('/v1.0/<string:tenant_id>/entities/<string:entity_id>/checks', methods=['POST'])
    def create_check(self, request, tenant_id, entity_id):
        """
        Create a check
        """
        content = request.content.read()
        postdata = json.loads(content)

        newcheck = None
        try:
            newcheck = create_check(self._session_store.clock, entity_id, postdata)
        except ValueError as err:
            match = MISSING_REQUIRED_KEY_REGEX.match(err.message)
            missing_key = match.group(1)
            status = 400
            request.setResponseCode(status)
            self._audit('checks', request, tenant_id, status, content)
            return json.dumps({'type': 'badRequest',
                               'code': status,
                               'message': 'Validation error for key \'{0}\''.format(missing_key),
                               'details': 'Missing required key ({0})'.format(missing_key),
                               'txnId': '.fake.mimic.transaction.id.c-1111111.ts-123444444.v-12344frf'})

        self._entity_cache_for_tenant(tenant_id).checks_list.append(newcheck)
        status = 201
        request.setResponseCode(status)
        request.setHeader('location', base_uri_from_request(request).rstrip('/') +
                          request.path + '/' + newcheck.id.encode('utf-8'))
        request.setHeader('x-object-id', newcheck.id.encode('utf-8'))
        request.setHeader('content-type', 'text/plain')
        self._audit('checks', request, tenant_id, status, content)
        return b''

    @app.route('/v1.0/<string:tenant_id>/entities/<string:entity_id>/checks/<string:check_id>',
               methods=['GET'])
    def get_check(self, request, tenant_id, entity_id, check_id):
        """
        Get a specific check that was created before
        """
        return _object_getter(self._entity_cache_for_tenant(tenant_id).checks_list,
                              lambda check: check.id == check_id,
                              request,
                              "Check",
                              '{0}:{1}'.format(entity_id, check_id))

    @app.route('/v1.0/<string:tenant_id>/entities/<string:entity_id>/checks/<string:check_id>',
               methods=['PUT'])
    def update_check(self, request, tenant_id, entity_id, check_id):
        """
        Updates a check in place.
        """
        content = request.content.read()
        update = json.loads(content)
        update_kwargs = dict(update)
        update_kwargs['clock'] = self._session_store.clock
        for check in self._entity_cache_for_tenant(tenant_id).checks_list:
            if check.entity_id == entity_id and check.id == check_id:
                check.update(**update_kwargs)
                break
        status = 204
        request.setResponseCode(status)
        request.setHeader('location', base_uri_from_request(request).rstrip('/') + request.path)
        request.setHeader('x-object-id', check_id.encode('utf-8'))
        request.setHeader('content-type', 'text/plain')
        self._audit('checks', request, tenant_id, status, content)
        return b''

    @app.route('/v1.0/<string:tenant_id>/entities/<string:entity_id>/checks/<string:check_id>',
               methods=['DELETE'])
    def delete_check(self, request, tenant_id, entity_id, check_id):
        """
        Deletes check and all alarms associated to it
        """
        checks = self._entity_cache_for_tenant(tenant_id).checks_list

        try:
            checks.remove(Matcher(
                lambda check: check.id == check_id and check.entity_id == entity_id))
        except ValueError:
            status = 404
            request.setResponseCode(status)
            self._audit('checks', request, tenant_id, status)
            return json.dumps({'type': 'notFoundError',
                               'code': status,
                               'message': 'Object does not exist',
                               'details': 'Object "Check" with key "{0}:{1}" does not exist'.format(
                                   entity_id, check_id),
                               'txnId': '.fake.mimic.transaction.id.c-1111111.ts-123444444.v-12344frf'})

        alarms = self._entity_cache_for_tenant(tenant_id).alarms_list
        alarms[:] = [alarm for alarm in alarms
                     if not (alarm.check_id == check_id and alarm.entity_id == entity_id)]
        status = 204
        request.setResponseCode(status)
        request.setHeader('content-type', 'text/plain')
        self._audit('checks', request, tenant_id, status)
        return b''

    @app.route('/v1.0/<string:tenant_id>/entities/<string:entity_id>/test-check', methods=['POST'])
    def test_check(self, request, tenant_id, entity_id):
        """
        Tests a check.

        If the user has configured overrides using the control API for
        test-check using this entity and check type, those will be used.
        Otherwise, random values within each metric type will be
        generated. For instance, integer metrics generate integers, and
        string metrics generate strings. No other guarantees are made.
        """
        content = request.content.read()
        test_config = json.loads(content)
        check_type = test_config['type']
        maas_store = self._entity_cache_for_tenant(tenant_id).maas_store
        response_code, response_body = maas_store.check_types[check_type].get_test_check_response(
            entity_id=entity_id,
            monitoring_zones=test_config.get('monitoring_zones_poll'),
            timestamp=int(1000 * self._session_store.clock.seconds()))
        request.setResponseCode(response_code)
        self._audit('checks', request, tenant_id, response_code, content)
        return json.dumps(response_body)

    @app.route('/v1.0/<string:tenant_id>/entities/<string:entity_id>/alarms', methods=['POST'])
    def create_alarm(self, request, tenant_id, entity_id):
        """
        Creates alarm
        """
        content = request.content.read()
        postdata = json.loads(content)

        try:
            newalarm = create_alarm(self._session_store.clock, entity_id, postdata)
        except ValueError as err:
            match = MISSING_REQUIRED_KEY_REGEX.match(err.message)
            missing_key = match.group(1)
            status = 400
            request.setResponseCode(status)
            self._audit('alarms', request, tenant_id, status, content)
            return json.dumps({'type': 'badRequest',
                               'code': status,
                               'message': 'Validation error for key \'{0}\''.format(missing_key),
                               'details': 'Missing required key ({0})'.format(missing_key),
                               'txnId': '.fake.mimic.transaction.id.c-1111111.ts-123444444.v-12344frf'})

        self._entity_cache_for_tenant(tenant_id).alarms_list.append(newalarm)
        status = 201
        request.setResponseCode(status)
        request.setHeader('location', base_uri_from_request(request).rstrip('/') +
                          request.path + '/' + newalarm.id.encode('utf-8'))
        request.setHeader('x-object-id', newalarm.id.encode('utf-8'))
        request.setHeader('content-type', 'text/plain')
        self._audit('alarms', request, tenant_id, status, content)
        return b''

    @app.route('/v1.0/<string:tenant_id>/entities/<string:entity_id>/alarms/<string:alarm_id>',
               methods=['GET'])
    def get_alarm(self, request, tenant_id, entity_id, alarm_id):
        """
        Gets an alarm by ID.
        """
        return _object_getter(self._entity_cache_for_tenant(tenant_id).alarms_list,
                              lambda alarm: alarm.entity_id == entity_id and alarm.id == alarm_id,
                              request,
                              "Alarm",
                              '{0}:{1}'.format(entity_id, alarm_id))

    @app.route('/v1.0/<string:tenant_id>/entities/<string:entity_id>/alarms/<string:alarm_id>',
               methods=['PUT'])
    def update_alarm(self, request, tenant_id, entity_id, alarm_id):
        """
        Updates an alarm in place.

        Documentation for this API can be found in the Rackspace Cloud
        Monitoring Developer Guide, section 5.12.5, "Update alarm by ID".
        The full link is quite long, but you can reach it by browsing
        to the following goo.gl URL:

            http://goo.gl/NhxgTZ
        """
        content = request.content.read()
        update = json.loads(content)
        update_kwargs = dict(update)
        update_kwargs['clock'] = self._session_store.clock
        for alarm in self._entity_cache_for_tenant(tenant_id).alarms_list:
            if alarm.entity_id == entity_id and alarm.id == alarm_id:
                alarm.update(**update_kwargs)
                break

        status = 204
        request.setResponseCode(status)
        request.setHeader('location', base_uri_from_request(request).rstrip('/') + request.path)
        request.setHeader('x-object-id', alarm_id.encode('utf-8'))
        request.setHeader('content-type', 'text/plain')
        self._audit('alarms', request, tenant_id, status, content)
        return b''

    @app.route('/v1.0/<string:tenant_id>/entities/<string:entity_id>/alarms/<string:alarm_id>',
               methods=['DELETE'])
    def delete_alarm(self, request, tenant_id, entity_id, alarm_id):
        """
        Delete an alarm
        """
        alarms = self._entity_cache_for_tenant(tenant_id).alarms_list

        try:
            alarms.remove(Matcher(
                lambda alarm: alarm.entity_id == entity_id and alarm.id == alarm_id))
        except ValueError:
            status = 404
            request.setResponseCode(status)
            self._audit('alarms', request, tenant_id, status)
            return json.dumps({'type': 'notFoundError',
                               'code': status,
                               'message': 'Object does not exist',
                               'details': 'Object "Alarm" with key "{0}:{1}" does not exist'.format(
                                   entity_id, alarm_id),
                               'txnId': '.fake.mimic.transaction.id.c-1111111.ts-123444444.v-12344frf'})

        status = 204
        request.setResponseCode(status)
        self._audit('alarms', request, tenant_id, status)
        request.setHeader('content-type', 'text/plain')
        return b''

    @app.route('/v1.0/<string:tenant_id>/entities/<string:entity_id>/test-alarm', methods=['POST'])
    def test_alarm(self, request, tenant_id, entity_id):
        """
        Test an alarm.

        This API can be driven using the control API to set an error
        or canned success response. If no error or success response is set,
        it will return success with a random state and status. Users should
        not expect this API to consistently return either OK, WARNING or
        CRITICAL without first setting the response in the control API.
        """
        content = request.content.read()
        payload = json.loads(content)
        n_tests = len(payload['check_data'])
        current_time_milliseconds = int(1000 * self._session_store.clock.seconds())
        status = 200
        response_payload = []

        test_responses = self._entity_cache_for_tenant(tenant_id).test_alarm_responses
        test_errors = self._entity_cache_for_tenant(tenant_id).test_alarm_errors

        if entity_id in test_errors and len(test_errors[entity_id]) > 0:
            error_response = test_errors[entity_id].popleft()
            status = error_response['code']
            response_payload = error_response['response']
        elif entity_id in test_responses:
            n_responses = len(test_responses[entity_id])
            for i in xrange(n_tests):
                test_response = test_responses[entity_id][i % n_responses]
                response_payload.append({'state': test_response['state'],
                                         'status': test_response.get(
                                             'status', 'Matched default return statement'),
                                         'timestamp': current_time_milliseconds})
        else:
            for _ in xrange(n_tests):
                response_payload.append({'state': random.choice(['OK', 'WARNING', 'CRITICAL']),
                                         'status': random_hipsum(12),
                                         'timestamp': current_time_milliseconds})

        request.setResponseCode(status)
        self._audit('alarms', request, tenant_id, status, content)
        return json.dumps(response_payload)

    @app.route('/v1.0/<string:tenant_id>/entities/<string:entity_id>/alarms', methods=['GET'])
    def get_alarms_for_entity(self, request, tenant_id, entity_id):
        """
        Get all alarms for the specified entity.
        """
        alarms = [alarm.to_json()
                  for alarm in self._entity_cache_for_tenant(tenant_id).alarms_list
                  if alarm.entity_id == entity_id]
        metadata = {'count': len(alarms),
                    'limit': 1000,
                    'marker': None,
                    'next_marker': None,
                    'next_href': None}
        request.setResponseCode(200)
        return json.dumps({'metadata': metadata, 'values': alarms})

    @app.route('/v1.0/<string:tenant_id>/views/overview', methods=['GET'])
    def overview(self, request, tenant_id):
        """
        serves the overview api call,returns all entities,checks and alarms
        """
        all_entities = self._entity_cache_for_tenant(tenant_id).entities_list
        if 'entityId' in request.args:
            all_entities = [entity for entity in all_entities
                            if entity.id in request.args['entityId']]
            if len(all_entities) == 0:
                request.setResponseCode(404)
                return json.dumps({'type': 'notFoundError',
                                   'code': 404,
                                   'message': 'Object does not exist',
                                   'details': 'Object "Entity" with key "{0}" does not exist'.format(
                                       request.args['entityId'])})

        checks = self._entity_cache_for_tenant(tenant_id).checks_list
        alarms = self._entity_cache_for_tenant(tenant_id).alarms_list
        maas_store = self._entity_cache_for_tenant(tenant_id).maas_store
        page_limit = min(int(request.args.get('limit', [100])[0]), 1000)
        offset = 0
        current_marker = request.args.get('marker', [None])[0]
        if current_marker is not None:
            try:
                offset = all_entities.index(Matcher(lambda entity: entity.id == current_marker))
            except ValueError:
                offset = 0

        entities = all_entities[offset:offset + page_limit]
        next_marker = None
        if offset + page_limit < len(all_entities):
            next_marker = all_entities[offset + page_limit].id

        metadata = {
            'count': len(entities),
            'marker': current_marker,
            'next_marker': next_marker,
            'limit': page_limit,
            'next_href': None
        }
        values = [{'alarms': [alarm.to_json()
                              for alarm in alarms
                              if alarm.entity_id == entity.id],
                   'checks': [check.to_json()
                              for check in checks
                              if check.entity_id == entity.id],
                   'entity': entity.to_json(),
                   'latest_alarm_states': [
                       state.brief_json()
                       for state in maas_store.latest_alarm_states_for_entity(entity.id)]}

                  for entity in entities]
        request.setResponseCode(200)
        return json.dumps({'metadata': metadata, 'values': values})

    @app.route('/v1.0/<string:tenant_id>/audits', methods=['GET'])
    def list_audits(self, request, tenant_id):
        """
        Gets the user's audit logs.
        """
        ordering = -1 if request.args.get('reverse', False) else 1
        all_audits = self._entity_cache_for_tenant(tenant_id).audits_list[::ordering]
        page_limit = min(int(request.args.get('limit', [100])[0]), 1000)
        offset = 0
        current_marker = request.args.get('marker', [None])[0]
        if current_marker is not None:
            try:
                offset = all_audits.index(Matcher(lambda audit: audit['id'] == current_marker))
            except ValueError:
                offset = 0

        audits = all_audits[offset:offset + page_limit]
        next_marker = None
        if offset + page_limit < len(all_audits):
            next_marker = all_audits[offset + page_limit]['id']

        metadata = {
            'count': len(audits),
            'marker': current_marker,
            'next_marker': next_marker,
            'limit': page_limit,
            'next_href': None
        }
        request.setResponseCode(200)
        return json.dumps({'metadata': metadata, 'values': audits})

    @app.route('/v1.0/<string:tenant_id>/__experiments/json_home', methods=['GET'])
    def service_json_home(self, request, tenant_id):
        """
        jsonhome call. CloudIntellgiences doesn't actually use these URLs directly.
        Rather, do some regex on them to figure how to know what permissions the user as
        have
        TO DO: Regionless api
        """
        request.setResponseCode(200)
        mockapi_id = re.findall('/mimicking/(.+?)/', request.path)[0]
        url = base_uri_from_request(request).rstrip('/') + '/mimicking/' + mockapi_id + '/ORD/v1.0'
        return json.dumps(json_home(url))

    @app.route('/v1.0/<string:tenant_id>/views/agent_host_info', methods=['GET'])
    def view_agent_host_info(self, request, tenant_id):
        """
        Return 400 until the fifth attempt, then start to return data as if the
        Agent is truly installed and working.
        """
        if 'entityId' not in request.args:
            request.setResponseCode(400)
            return json.dumps({'type': 'badRequest',
                               'code': 400,
                               'message': 'Validation error for key \'agentId, entityId, uri\'',
                               'details': 'You must specify an agentId, entityId, or an entity URI.',
                               'mimicNotes': 'But mimic will only accept entityId right now',
                               'txnId': '.fake.mimic.transaction.id.c-1111111.ts-123444444.v-12344frf'})

        entity_id = request.args['entityId'][0].strip()
        for e in self._entity_cache_for_tenant(tenant_id).entities_list:
            if e.id == entity_id:
                if e.agent_id is None:
                    e.agent_id = random_hex_generator(12)
                agent_id = e.agent_id
                break
        else:
            request.setResponseCode(404)
            return json.dumps({'type': 'notFoundError',
                               'code': 404,
                               'message': 'Object does not exist',
                               'details': 'Object "Entity" with key "{0}" does not exist'.format(
                                   entity_id),
                               'txnId': '.fake.mimic.transaction.id.c-1111111.ts-123444444.v-12344frf'
                               })

        self._entity_cache_for_tenant(tenant_id).agenthostinfo_querycount[entity_id] += 1
        if self._entity_cache_for_tenant(tenant_id).agenthostinfo_querycount[entity_id] < 5:
            request.setResponseCode(400)
            return json.dumps({
                "type": "agentDoesNotExist",
                "code": 400,
                "message": "Agent does not exist",
                "details": "Agent XYZ does not exist.",
                "txnId": ".fake.mimic.transaction.id.c-1111111.ts-123444444.v-12344frf"
            })
        else:
            return json.dumps(agent_info(entity_id, agent_id))

    @app.route('/v1.0/<string:tenant_id>/agent_installers', methods=['POST'])
    def agent_installer(self, request, tenant_id):
        """
        URL of agent install script
        """
        xsil = "https://monitoring.api.rackspacecloud.com/"
        xsil += "v1.0/00000/agent_installers/c69b2ceafc0444506fb32255af3d9be3.sh"
        status = 201
        request.setResponseCode(status)
        request.setHeader('x-shell-installer-location', xsil)
        self._audit('agent_installers', request, tenant_id, status, request.content.read())
        return b''

    @app.route('/v1.0/<string:tenant_id>/notifications', methods=['POST'])
    def create_notification(self, request, tenant_id):
        """
        Create notification target
        """
        content = request.content.read()
        new_n = create_notification(self._session_store.clock, json.loads(content))
        self._entity_cache_for_tenant(tenant_id).notifications_list.append(new_n)
        status = 201
        request.setResponseCode(status)
        request.setHeader('content-type', 'text/plain')
        request.setHeader('location', base_uri_from_request(request).rstrip('/') +
                          request.path + '/' + new_n.id.encode('utf-8'))
        request.setHeader('x-object-id', new_n.id.encode('utf-8'))
        self._audit('notifications', request, tenant_id, status, content)
        return b''

    @app.route('/v1.0/<string:tenant_id>/notifications', methods=['GET'])
    def get_notifications(self, request, tenant_id):
        """
        Get notification targets
        """
        nt_list = self._entity_cache_for_tenant(tenant_id).notifications_list
        metadata = {'count': len(nt_list),
                    'limit': 100,
                    'marker': None,
                    'next_marker': None,
                    'next_href': None}
        request.setResponseCode(200)
        return json.dumps({'values': [nt.to_json() for nt in nt_list], 'metadata': metadata})

    @app.route('/v1.0/<string:tenant_id>/notifications/<string:nt_id>', methods=['PUT'])
    def update_notifications(self, request, tenant_id, nt_id):
        """
        Updates notification targets
        """
        content = request.content.read()
        postdata = json.loads(content)
        update_kwargs = dict(postdata)
        update_kwargs['clock'] = self._session_store.clock
        nt_list = self._entity_cache_for_tenant(tenant_id).notifications_list
        for nt in nt_list:
            if nt.id == nt_id:
                nt.update(**update_kwargs)
                break

        status = 204
        request.setResponseCode(status)
        request.setHeader('content-type', 'text/plain')
        self._audit('notifications', request, tenant_id, status, content)
        return b''

    @app.route('/v1.0/<string:tenant_id>/notifications/<string:nt_id>', methods=['DELETE'])
    def delete_notification(self, request, tenant_id, nt_id):
        """
        Delete a notification
        """
        notifications = self._entity_cache_for_tenant(tenant_id).notifications_list

        try:
            notifications.remove(Matcher(lambda nt: nt.id == nt_id))
        except ValueError:
            status = 404
            request.setResponseCode(status)
            self._audit('notifications', request, tenant_id, status)
            return json.dumps({'type': 'notFoundError',
                               'code': status,
                               'message': 'Object does not exist',
                               'details': 'Object "Notification" with key "{0}" does not exist'.format(
                                   nt_id),
                               'txnId': '.fake.mimic.transaction.id.c-1111111.ts-123444444.v-12344frf'})

        status = 204
        request.setResponseCode(status)
        self._audit('notifications', request, tenant_id, status)
        request.setHeader('content-type', 'text/plain')
        return b''

    @app.route('/v1.0/<string:tenant_id>/notification_plans', methods=['POST'])
    def create_notificationplan(self, request, tenant_id):
        """
        Creates a new notificationPlans
        """
        content = request.content.read()
        postdata = json.loads(content)
        newnp = create_notification_plan(self._session_store.clock, postdata)
        self._entity_cache_for_tenant(tenant_id).notificationplans_list.append(newnp)
        status = 201
        request.setResponseCode(status)
        request.setHeader('content-type', 'text/plain')
        request.setHeader('location', base_uri_from_request(request).rstrip('/') +
                          request.path + '/' + newnp.id.encode('utf-8'))
        request.setHeader('x-object-id', newnp.id.encode('utf-8'))
        self._audit('notification_plans', request, tenant_id, status, content)
        return b''

    @app.route('/v1.0/<string:tenant_id>/notification_plans', methods=['GET'])
    def get_notification_plans(self, request, tenant_id):
        """
        Get all notification plans
        """
        np_list = self._entity_cache_for_tenant(tenant_id).notificationplans_list
        metadata = {'count': len(np_list),
                    'limit': 100,
                    'marker': None,
                    'next_marker': None,
                    'next_href': None}
        request.setResponseCode(200)
        return json.dumps({'values': [np.to_json() for np in np_list], 'metadata': metadata})

    @app.route('/v1.0/<string:tenant_id>/notification_plans/<string:np_id>', methods=['GET'])
    def get_notification_plan(self, request, tenant_id, np_id):
        """
        Get specific notif plan
        """
        return _object_getter(self._entity_cache_for_tenant(tenant_id).notificationplans_list,
                              lambda np: np.id == np_id,
                              request,
                              "Notification Plan",
                              np_id)

    @app.route('/v1.0/<string:tenant_id>/notification_plans/<string:np_id>', methods=['PUT'])
    def update_notification_plan(self, request, tenant_id, np_id):
        """
        Alter a notification plan
        """
        content = request.content.read()
        postdata = json.loads(content)
        update_kwargs = dict(postdata)
        update_kwargs['clock'] = self._session_store.clock
        np_list = self._entity_cache_for_tenant(tenant_id).notificationplans_list
        for np in np_list:
            if np.id == np_id:
                np.update(**update_kwargs)
                break
        status = 204
        request.setResponseCode(status)
        request.setHeader('content-type', 'text/plain')
        self._audit('notification_plans', request, tenant_id, status, content)
        return b''

    @app.route('/v1.0/<string:tenant_id>/notification_plans/<string:np_id>', methods=['DELETE'])
    def delete_notification_plan(self, request, tenant_id, np_id):
        """
        Remove a notification plan
        """
        all_alarms = self._entity_cache_for_tenant(tenant_id).alarms_list
        nplist = self._entity_cache_for_tenant(tenant_id).notificationplans_list
        alarmids_using_np = [alarm.id for alarm in all_alarms
                             if alarm.notification_plan_id == np_id]

        if len(alarmids_using_np):
            status = 403
            request.setResponseCode(status)
            err_message = ('Notification plans cannot be removed while alarms ' +
                           'are using it: {0}'.format(' '.join(alarmids_using_np)))
            self._audit('notification_plans', request, tenant_id, status)
            return json.dumps({'type': 'forbiddenError',
                               'code': status,
                               'txnId': '.fake.mimic.transaction.id.c-1111111.ts-123444444.v-12344frf',
                               'message': err_message,
                               'details': err_message})

        try:
            nplist.remove(Matcher(lambda np: np.id == np_id))
        except ValueError:
            status = 404
            request.setResponseCode(status)
            self._audit('notification_plans', request, tenant_id, status)
            return json.dumps({'type': 'notFoundError',
                               'code': status,
                               'message': 'Object does not exist',
                               'details': ('Object "NotificationPlan" with key "{0}" '.format(np_id) +
                                           'does not exist'),
                               'txnId': '.fake.mimic.transaction.id.c-1111111.ts-123444444.v-12344frf'})

        status = 204
        request.setResponseCode(status)
        self._audit('notification_plans', request, tenant_id, status)
        request.setHeader('content-type', 'text/plain')
        return b''

    @app.route('/v1.0/<string:tenant_id>/suppressions', methods=['GET'])
    def get_suppressions(self, request, tenant_id):
        """
        Get the list of suppressions for this tenant.
        """
        sp_list = self._entity_cache_for_tenant(tenant_id).suppressions_list
        metadata = {
            'count': len(sp_list),
            'limit': 100,
            'marker': None,
            'next_marker': None,
            'next_href': None
        }
        request.setResponseCode(200)
        return json.dumps({'values': [sp.to_json() for sp in sp_list], 'metadata': metadata})

    @app.route('/v1.0/<string:tenant_id>/suppressions/<string:sp_id>', methods=['GET'])
    def get_suppression(self, request, tenant_id, sp_id):
        """
        Get a suppression by ID.
        """
        return _object_getter(self._entity_cache_for_tenant(tenant_id).suppressions_list,
                              lambda sp: sp.id == sp_id,
                              request,
                              "Suppression",
                              sp_id)

    @app.route('/v1.0/<string:tenant_id>/suppressions', methods=['POST'])
    def create_suppression(self, request, tenant_id):
        """
        Create a new suppression.
        """
        content = request.content.read()
        postdata = json.loads(content)
        newsp = create_suppression(self._session_store.clock, postdata)
        self._entity_cache_for_tenant(tenant_id).suppressions_list.append(newsp)
        status = 201
        request.setResponseCode(status)
        request.setHeader('location', base_uri_from_request(request).rstrip('/') +
                          request.path + '/' + newsp.id.encode('utf-8'))
        request.setHeader('x-object-id', newsp.id.encode('utf-8'))
        request.setHeader('content-type', 'text/plain')
        self._audit('suppressions', request, tenant_id, status, content)
        return b''

    @app.route('/v1.0/<string:tenant_id>/suppressions/<string:sp_id>', methods=['PUT'])
    def update_suppression(self, request, tenant_id, sp_id):
        """
        Update a suppression.
        """
        content = request.content.read()
        postdata = json.loads(content)
        update_kwargs = dict(postdata)
        update_kwargs['clock'] = self._session_store.clock
        sp_list = self._entity_cache_for_tenant(tenant_id).suppressions_list
        for sp in sp_list:
            if sp.id == sp_id:
                sp.update(**update_kwargs)
                break
        status = 204
        request.setResponseCode(status)
        request.setHeader('content-type', 'text/plain')
        self._audit('suppressions', request, tenant_id, status, content)
        return b''

    @app.route('/v1.0/<string:tenant_id>/suppressions/<string:sp_id>', methods=['DELETE'])
    def delete_suppression(self, request, tenant_id, sp_id):
        """
        Delete a suppression.
        """
        suppressions = self._entity_cache_for_tenant(tenant_id).suppressions_list

        try:
            suppressions.remove(Matcher(lambda sp: sp.id == sp_id))
        except ValueError:
            status = 404
            request.setResponseCode(status)
            self._audit('suppressions', request, tenant_id, status)
            return json.dumps({'type': 'notFoundError',
                               'code': status,
                               'message': 'Object does not exist',
                               'details': 'Object "Suppression" with key "{0}" does not exist'.format(
                                   sp_id),
                               'txnId': '.fake.mimic.transaction.id.c-1111111.ts-123444444.v-12344frf'})

        status = 204
        request.setResponseCode(status)
        self._audit('suppressions', request, tenant_id, status)
        request.setHeader('content-type', 'text/plain')
        return b''

    @app.route('/v1.0/<string:tenant_id>/monitoring_zones', methods=['GET'])
    def list_monitoring_zones(self, request, tenant_id):
        """
        Lists the monitoring zones
        """
        mzs = monitoring_zones()
        metadata = {
            'count': len(mzs),
            'limit': 100,
            'marker': None,
            'next_marker': None,
            'next_href': None
        }
        request.setResponseCode(200)
        return json.dumps({'values': mzs, 'metadata': metadata})

    @app.route('/v1.0/<string:tenant_id>/alarm_examples', methods=['GET'])
    def list_alarm_examples(self, request, tenant_id):
        """
        Lists all of the alarm examples.
        """
        axs = alarm_examples()
        metadata = {
            'count': len(axs),
            'limit': 100,
            'marker': None,
            'next_marker': None,
            'next_href': None
        }
        request.setResponseCode(200)
        return json.dumps({'values': axs, 'metadata': metadata})

    @app.route('/v1.0/<string:tenant_id>/views/alarmCountsPerNp', methods=['GET'])
    def alarm_counts_per_np(self, request, tenant_id):
        """
        All NotificationPlans a number of alarms pointing to them.
        """
        all_alarms = self._entity_cache_for_tenant(tenant_id).alarms_list
        all_nps = self._entity_cache_for_tenant(tenant_id).notificationplans_list

        values = [{'notification_plan_id': np.id,
                   'alarm_count': len([alarm for alarm in all_alarms
                                       if alarm.notification_plan_id == np.id])}
                  for np in all_nps]

        metadata = {'limit': 100,
                    'marker': None,
                    'next_marker': None,
                    'next_href': None,
                    'count': len(values)}
        request.setResponseCode(200)
        return json.dumps({'values': values, 'metadata': metadata})

    @app.route('/v1.0/<string:tenant_id>/views/alarmsByNp/<string:np_id>', methods=['GET'])
    def alarms_by_np(self, request, tenant_id, np_id):
        """
        List of alarms pointing to a particular NotificationPlan
        """
        all_alarms = self._entity_cache_for_tenant(tenant_id).alarms_list
        values = [alarm.to_json() for alarm in all_alarms
                  if alarm.notification_plan_id == np_id]
        metadata = {'limit': 100,
                    'marker': None,
                    'next_marker': None,
                    'next_href': None,
                    'count': len(values)}
        request.setResponseCode(200)
        return json.dumps({'values': values, 'metadata': metadata})

    @app.route('/v1.0/<string:tenant_id>/notification_types', methods=['GET'])
    def get_notification_types(self, request, tenant_id):
        """
        Get the types of notifications supported: pageduty,email,sms, etc
        """
        ntlist = self._entity_cache_for_tenant(tenant_id).notificationtypes_list
        metadata = {'count': len(ntlist),
                    'limit': 100,
                    'marker': None,
                    'next_marker': None,
                    'next_href': None}
        request.setResponseCode(200)
        return json.dumps({'values': ntlist, 'metadata': metadata})

    @app.route('/v1.0/<string:tenant_id>/views/metric_list', methods=['GET'])
    def views_metric_list(self, request, tenant_id):
        """
        All available metrics.
        """
        entities = self._entity_cache_for_tenant(tenant_id).entities_list
        all_checks = self._entity_cache_for_tenant(tenant_id).checks_list
        maas_store = self._entity_cache_for_tenant(tenant_id).maas_store
        values = [_metric_list_for_entity(maas_store, entity, all_checks)
                  for entity in entities]

        metadata = {'count': len(values),
                    'marker': None,
                    'next_marker': None,
                    'limit': 1000,
                    'next_href': None}

        request.setResponseCode(200)
        return json.dumps({'metadata': metadata, 'values': values})

    @app.route('/v1.0/<string:tenant_id>/__experiments/multiplot', methods=['POST'])
    def multiplot(self, request, tenant_id):
        """
        datapoints for all metrics requested
        Right now, only checks of type remote.ping work
        """
        checks = self._entity_cache_for_tenant(tenant_id).checks_list
        maas_store = self._entity_cache_for_tenant(tenant_id).maas_store
        content = request.content.read()
        multiplot_request = json.loads(content)

        requested_check_ids = set([metric['check_id'] for metric in multiplot_request['metrics']])
        checks_by_id = dict([(check.id, check)
                             for check in checks
                             if check.id in requested_check_ids])

        for requested_metric in multiplot_request['metrics']:
            if requested_metric['check_id'] not in checks_by_id:
                status = 400
                request.setResponseCode(status)
                self._audit('rollups', request, tenant_id, status, content)
                return json.dumps({
                    'type': 'requiredNotFoundError',
                    'code': status,
                    'message': 'Required object does not exist',
                    'details': 'Object "Check" with key "{0},{1}" does not exist'.format(
                        requested_metric['entity_id'], requested_metric['check_id']),
                    'txnId': '.fake.mimic.transaction.id.c-1111111.ts-123444444.v-12344frf'})

        multiplot_metrics = [_compute_multiplot(maas_store,
                                                metric['entity_id'],
                                                checks_by_id[metric['check_id']],
                                                metric['metric'],
                                                int(request.args['from'][0]),
                                                int(request.args['to'][0]),
                                                int(request.args['points'][0]))
                             for metric in multiplot_request['metrics']]
        status = 200
        request.setResponseCode(200)
        self._audit('rollups', request, tenant_id, status, content)
        return json.dumps({'metrics': multiplot_metrics})

    @app.route('/v1.0/<string:tenant_id>/views/latest_alarm_states', methods=['GET'])
    def latest_alarm_states(self, request, tenant_id):
        """
        Gets entities grouped with their latest alarm states.
        """
        entities = self._entity_cache_for_tenant(tenant_id).entities_list
        maas_store = self._entity_cache_for_tenant(tenant_id).maas_store

        values = [{'entity_id': entity.id,
                   'entity_uri': entity.uri,
                   'entity_label': entity.label,
                   'latest_alarm_states': [
                       state.detail_json()
                       for state in maas_store.latest_alarm_states_for_entity(entity.id)]}
                  for entity in entities]

        metadata = {'count': len(values),
                    'marker': None,
                    'next_marker': None,
                    'limit': 1000,
                    'next_href': None}

        request.setResponseCode(200)
        return json.dumps({'values': values, 'metadata': metadata})


@implementer(IAPIMock, IPlugin)
@attributes(["maas_api"])
class MaasControlApi(object):
    """
    This class registers the MaaS controller API in the service catalog.
    """
    def catalog_entries(self, tenant_id):
        """
        List catalog entries for the MaaS API.
        """
        return [
            Entry(
                tenant_id, "rax: monitor", "cloudMonitoringControl",
                [
                    Endpoint(tenant_id, region, text_type(uuid4()),
                             "v1.0")
                    for region in self.maas_api._regions
                ]
            )
        ]

    def resource_for_region(self, region, uri_prefix, session_store):
        """
        Get an :obj:`twisted.web.iweb.IResource` for the given URI prefix;
        implement :obj:`IAPIMock`.
        """
        maas_controller = MaasController(api_mock=self,
                                         session_store=session_store,
                                         region=region)
        return maas_controller.app.resource()


@attributes(["api_mock", "session_store", "region"])
class MaasController(object):
    """
    Klein routes for MaaS control API.
    """

    def _entity_cache_for_tenant(self, tenant_id):
        """
        Retrieve the M_cache object containing all objects created so far
        """
        clock = self.session_store.clock
        return (self.session_store.session_for_tenant_id(tenant_id)
                .data_for_api(self.api_mock.maas_api, _mcache_factory(clock))[self.region])

    app = MimicApp()

    @app.route('/v1.0/<string:tenant_id>/entities/<string:entity_id>/alarms/test_response',
               methods=['PUT'])
    def set_test_alarm_response(self, request, tenant_id, entity_id):
        """
        Sets the test-alarm response for a given entity.
        """
        test_responses = self._entity_cache_for_tenant(tenant_id).test_alarm_responses
        dummy_response = json.loads(request.content.read())
        test_responses[entity_id] = []
        for response_block in dummy_response:
            ith_response = {'state': response_block['state']}
            if 'status' in response_block:
                ith_response['status'] = response_block['status']
            test_responses[entity_id].append(ith_response)
        request.setResponseCode(204)
        return b''

    @app.route('/v1.0/<string:tenant_id>/entities/<string:entity_id>/alarms/test_errors',
               methods=['POST'])
    def push_test_alarm_error(self, request, tenant_id, entity_id):
        """
        Creates a new error response that will be returned from the
        test-alarm API the next time it is called for this entity.
        """
        test_alarm_errors = self._entity_cache_for_tenant(tenant_id).test_alarm_errors
        request_body = json.loads(request.content.read())

        if entity_id not in test_alarm_errors:
            test_alarm_errors[entity_id] = collections.deque()

        error_obj = {'id': 'er' + random_hex_generator(4),
                     'code': request_body['code'],
                     'response': request_body['response']}
        test_alarm_errors[entity_id].append(error_obj)
        request.setResponseCode(201)
        request.setHeader('x-object-id', error_obj['id'])
        return b''

    @app.route('/v1.0/<string:tenant_id>/entities/<string:entity_id>/alarms/test_response',
               methods=['DELETE'])
    def clear_test_alarm_response(self, request, tenant_id, entity_id):
        """
        Clears the test-alarm response and restores normal behavior.
        """
        test_responses = self._entity_cache_for_tenant(tenant_id).test_alarm_responses
        del test_responses[entity_id]
        request.setResponseCode(204)
        return b''

    @app.route('/v1.0/<string:tenant_id>/entities/<string:entity_id>/checks' +
               '/test_responses/<string:check_type>', methods=['PUT'])
    def set_test_check_overrides(self, request, tenant_id, entity_id, check_type):
        """
        Sets overriding behavior on the test-check handler for a given
        entity ID and check type.
        """
        maas_store = self._entity_cache_for_tenant(tenant_id).maas_store
        check_type_ins = maas_store.check_types[check_type]
        overrides = json.loads(request.content.read())
        check_id = '__test_check'
        ench_key = (entity_id, check_id)

        for override in overrides:
            if 'available' in override:
                check_type_ins.test_check_available[ench_key] = override['available']
            if 'status' in override:
                check_type_ins.test_check_status[ench_key] = override['status']
            metrics_dict = override.get('metrics', {})
            for metric_name in metrics_dict:
                test_check_metric = check_type_ins.get_metric_by_name(metric_name)
                kwargs = {'entity_id': entity_id,
                          'check_id': check_id,
                          'override_fn': lambda _: metrics_dict[metric_name]['data']}
                if 'monitoring_zone_id' in override:
                    kwargs['monitoring_zone'] = override['monitoring_zone_id']
                test_check_metric.set_override(**kwargs)

        request.setResponseCode(204)
        return b''

    @app.route('/v1.0/<string:tenant_id>/entities/<string:entity_id>/checks' +
               '/test_responses/<string:check_type>', methods=['DELETE'])
    def clear_test_check_overrides(self, request, tenant_id, entity_id, check_type):
        """
        Clears overriding behavior on a test-check handler.
        """
        maas_store = self._entity_cache_for_tenant(tenant_id).maas_store
        check_type_ins = maas_store.check_types[check_type]
        check_type_ins.clear_overrides()
        request.setResponseCode(204)
        return b''

    @app.route('/v1.0/<string:tenant_id>/entities/<string:entity_id>/alarms' +
               '/<string:alarm_id>/states', methods=['POST'])
    def create_alarm_state(self, request, tenant_id, entity_id, alarm_id):
        """
        Adds a new alarm state to the collection of alarm states.
        """
        maas_store = self._entity_cache_for_tenant(tenant_id).maas_store
        request_body = json.loads(request.content.read())

        check_id = None
        alarm_label = None
        for al in self._entity_cache_for_tenant(tenant_id).alarms_list:
            if al.entity_id == entity_id and al.id == alarm_id:
                alarm_label = al.label
                check_id = al.check_id
                break
        else:
            request.setResponseCode(404)
            return json.dumps({'type': 'notFoundError',
                               'code': 404,
                               'message': 'Object does not exist',
                               'details': 'Object "Alarm" with key "{0}:{1}" does not exist'.format(
                                   entity_id, alarm_id)
                               })

        previous_state = u'UNKNOWN'
        alarm_states_same_entity_and_alarm = [
            state for state in maas_store.alarm_states
            if state.entity_id == entity_id and state.alarm_id == alarm_id]
        if len(alarm_states_same_entity_and_alarm) > 0:
            previous_state = alarm_states_same_entity_and_alarm[-1].state

        monitoring_zone_id = request_body.get('analyzed_by_monitoring_zone_id', u'mzord')

        new_state = None
        try:
            new_state = AlarmState(alarm_id=alarm_id,
                                   entity_id=entity_id,
                                   check_id=check_id,
                                   alarm_label=alarm_label,
                                   analyzed_by_monitoring_zone_id=monitoring_zone_id,
                                   previous_state=previous_state,
                                   state=request_body['state'],
                                   status=request_body['status'],
                                   timestamp=int(1000 * self.session_store.clock.seconds()))
        except KeyError as e:
            missing_key = e.message
            status = 400
            request.setResponseCode(status)
            return json.dumps({'type': 'badRequest',
                               'code': status,
                               'message': 'Validation error for key \'{0}\''.format(missing_key),
                               'details': 'Missing required key ({0})'.format(missing_key),
                               'txnId': '.fake.mimic.transaction.id.c-1111111.ts-123444444.v-12344frf'})

        maas_store.alarm_states.append(new_state)
        request.setResponseCode(201)
        request.setHeader('x-object-id', new_state.id.encode('utf-8'))
        request.setHeader('content-type', 'text/plain')
        return b''

    @app.route('/v1.0/<string:tenant_id>/entities/<string:entity_id>/checks' +
               '/<string:check_id>/metrics/<string:metric_name>', methods=['PUT'])
    def set_metric_override(self, request, tenant_id, entity_id, check_id, metric_name):
        """
        Sets overrides on a metric.
        """
        checks = self._entity_cache_for_tenant(tenant_id).checks_list
        check = None

        for c in checks:
            if c.id == check_id and c.entity_id == entity_id:
                check = c
                break
        else:
            request.setResponseCode(404)
            return json.dumps({'type': 'notFoundError',
                               'code': 404,
                               'message': 'Object does not exist',
                               'details': 'Object "Check" with key "{0}:{1}" does not exist'.format(
                                   entity_id, check_id)})

        maas_store = self._entity_cache_for_tenant(tenant_id).maas_store
        metric = maas_store.check_types[check.type].get_metric_by_name(metric_name)
        request_body = json.loads(request.content.read())
        monitoring_zones = request_body.get('monitoring_zones', ['__AGENT__'])
        override_type = request_body['type']
        override_options = request_body.get('options', {})
        override_fn = None

        if override_type == 'squarewave':
            fn_period = int(override_options.get('period', 10 * 60 * 1000))
            half_period = fn_period / 2
            fn_min = override_options.get('min', 20)
            fn_max = override_options.get('max', 80)
            fn_offset = int(override_options.get('offset', 0))
            override_fn = (lambda t: (fn_min
                                      if ((t + fn_offset) % fn_period) < half_period
                                      else fn_max))
        else:
            request.setResponseCode(400)
            return json.dumps({'type': 'badRequest',
                               'code': 400,
                               'message': 'Validation error for key \'type\'',
                               'details': 'Unknown value for "type": "{0}"'.format(override_type)})

        for monitoring_zone in monitoring_zones:
            metric.set_override(
                entity_id=entity_id,
                check_id=check_id,
                monitoring_zone=monitoring_zone,
                override_fn=override_fn)
        request.setResponseCode(204)
        return b''



"""
Model objects for images.
"""

from characteristic import attributes, Attribute
from json import dumps
from mimic.model.rackspace_images import (ImageStore, OnMetalImage)

from mimic.model.nova_objects import not_found
from mimic.canned_responses.mimic_presets import get_presets


@attributes(
    ["tenant_id", "region_name", "clock"])
class RegionalImageCollection(object):
    """
    A collection of images, in a given region, for a given tenant.
    """
    def image_by_id(self, image_id):
        """
        Retrieve a :obj:`Image` object by its ID.
        """
        images_store = ImageStore.create_image_store(self.tenant_id)
        for image in images_store:
            if image.image_id == image_id:
                return image

    def list_images(self, include_details, absolutize_url):
        """
        Return a list of images.
        """
        images_store = ImageStore.create_image_store(self.tenant_id)
        images = []
        for image in images_store:
            if self.region_name != "IAD" and isinstance(image, OnMetalImage):
                continue
            if include_details:
                images.append(image.detailed_json(absolutize_url))
            else:
                images.append(image.brief_json(absolutize_url))
        result = {"images": images}
        return dumps(result)

    def get_image(self, http_get_request, image_id, absolutize_url):
        """
        Return an image object if one exists from the list `/images` api,
        else return 404 Image not found.
        """
        if image_id in get_presets['servers']['invalid_image_ref'] or image_id.endswith('Z'):
            return dumps(not_found("The resource could not be found.",
                                   http_get_request))
        ImageStore.create_image_store(self.tenant_id)
        image = self.image_by_id(image_id)
        if image is None:
            return dumps(not_found('Image not found.', http_get_request))
        return dumps({"image": image.detailed_json(absolutize_url)})


@attributes(["tenant_id", "clock",
             Attribute("regional_collections", default_factory=dict)])
class GlobalImageCollection(object):
    """
    A :obj:`GlobalImageCollection` is a set of all the
    :obj:`RegionalImageCollection` objects owned by a given tenant.  In other
    words, all the image objects that a single tenant owns globally.
    """

    def collection_for_region(self, region_name):
        """
        Get a :obj:`RegionalFlavorCollection` for the region identified by the
        given name.
        """
        if region_name not in self.regional_collections:
            self.regional_collections[region_name] = (
                RegionalImageCollection(tenant_id=self.tenant_id, region_name=region_name,
                                        clock=self.clock))
        return self.regional_collections[region_name]



from mimic.rest.cloudfeeds import (CloudFeedsApi, CloudFeedsControlApi)
from mimic.test.fixtures import APIMockHelper
from mimic.test.helpers import request
from twisted.trial.unittest import SynchronousTestCase


class TestCloudFeedsAPI(SynchronousTestCase):
    def setUp(self):
        """
        Create a MimicCore with CloudFeeds support as the only plugin.
        """
        cf_api = CloudFeedsApi()
        self.helper = APIMockHelper(self, [cf_api, CloudFeedsControlApi(cf_api=cf_api)])
        self.root = self.helper.root
        self.uri = self.helper.uri
        self.ctrl_uri = self.helper.auth.get_service_endpoint(
            "cloudFeedsControl", "ORD"
        )

    def test_data_plane_access_should_404(self):
        """
        Attempts to hit the data plane, as it's currently written, should 404.
        This is because the boilerplate is written but no endpoints.
        """
        r = request(
            self, self.root, "GET", self.uri,
        )
        resp = self.successResultOf(r)
        self.assertEquals(resp.code, 404)

    def test_control_plane_access_should_404(self):
        """
        Attempts to hit the controldata plane, as it's currently written,
        should 404.  This is because the boilerplate is written but no
        endpoints.
        """
        r = request(
            self, self.root, "GET", self.ctrl_uri,
        )
        resp = self.successResultOf(r)
        self.assertEquals(resp.code, 404)



"""
Model objects for the Nova mimic.
"""

import re

from characteristic import attributes, Attribute
from random import randrange
from json import loads, dumps
from urllib import urlencode

from six import string_types

from mimic.util.helper import (
    seconds_to_timestamp,
    random_string,
    timestamp_to_seconds
)

from mimic.model.behaviors import (
    BehaviorRegistryCollection, EventDescription, Criterion, regexp_predicate
)
from twisted.web.http import ACCEPTED, BAD_REQUEST, FORBIDDEN, NOT_FOUND, CONFLICT


@attributes(['nova_message'])
class LimitError(Exception):
    """
    Error to be raised when a limit has been exceeded.
    """


@attributes(['nova_message'])
class BadRequestError(Exception):
    """
    Error to be raised when bad input has been received to Nova.
    """


def _nova_error_message(msg_type, message, status_code, request):
    """
    Set the response code on the request, and return a JSON blob representing
    a Nova error body, in the format Nova returns error messages.

    :param str msg_type: What type of error this is - something like
        "badRequest" or "itemNotFound" for Nova.
    :param str message: The message to include in the body.
    :param int status_code: The status code to set
    :param request: the request to set the status code on

    :return: dictionary representing the error body
    """
    request.setResponseCode(status_code)
    return {
        msg_type: {
            "message": message,
            "code": status_code
        }
    }


def bad_request(message, request):
    """
    Return a 400 error body associated with a Nova bad request error.
    Also sets the response code on the request.

    :param str message: The message to include in the bad request body.
    :param request: The request on which to set the response code.

    :return: dictionary representing the error body.
    """
    return _nova_error_message("badRequest", message, BAD_REQUEST, request)


def not_found(message, request):
    """
    Return a 404 error body associated with a Nova not found error.
    Also sets the response code on the request.

    :param str message: The message to include in the bad request body.
    :param request: The request on which to set the response code.

    :return: dictionary representing the error body.
    """
    return _nova_error_message("itemNotFound", message, NOT_FOUND, request)


def forbidden(message, request):
    """
    Return a 403 error body associated with a Nova forbidden error.
    Also sets the response code on the request.

    :param str message: The message to include in the bad request body.
    :param request: The request on which to set the response code.

    :return: dictionary representing the error body.
    """
    return _nova_error_message("forbidden", message, FORBIDDEN, request)


def conflicting(message, request):
    """
    Return a 409 error body associated with a Nova conflicting request error.

    :param str message: The message to include in the bad request body.
    :param request: The request on which to set the response code.

    :return: dictionary representing the error body.
    """
    return _nova_error_message("conflictingRequest", message, CONFLICT, request)


@attributes(["collection", "server_id", "server_name", "metadata",
             "creation_time", "update_time", "public_ips", "private_ips",
             "status", "flavor_ref", "image_ref", "disk_config", "key_name",
             "admin_password", "creation_request_json",
             Attribute('max_metadata_items', instance_of=int,
                       default_value=40)])
class Server(object):
    """
    A :obj:`Server` is a representation of all the state associated with a nova
    server.  It can produce JSON-serializable objects for various pieces of
    state that are required for API responses.
    """

    static_defaults = {
        "OS-EXT-STS:power_state": 1,
        "OS-EXT-STS:task_state": None,
        "accessIPv4": "198.101.241.238",  # TODO: same as public IP
        "accessIPv6": "2001:4800:780e:0510:d87b:9cbc:ff04:513a",
        "hostId": "33ccb6c82f3625748b6f2338f54d8e9df07cc583251e001355569056",
        "progress": 100,
        "user_id": "170454"
    }

    def addresses_json(self):
        """
        Create a JSON-serializable data structure describing the public and
        private IPs associated with this server.
        """
        return {
            "private": [addr.json() for addr in self.private_ips],
            "public": [addr.json() for addr in self.public_ips]
        }

    def links_json(self, absolutize_url):
        """
        Create a JSON-serializable data structure describing the links to this
        server.

        :param callable absolutize_url: see :obj:`default_create_behavior`.
        """
        tenant_id = self.collection.tenant_id
        server_id = self.server_id
        return [
            {
                "href": absolutize_url("v2/{0}/servers/{1}"
                                       .format(tenant_id, server_id)),
                "rel": "self"
            },
            {
                "href": absolutize_url("{0}/servers/{1}"
                                       .format(tenant_id, server_id)),
                "rel": "bookmark"
            }
        ]

    def brief_json(self, absolutize_url):
        """
        Brief JSON-serializable version of this server, for the non-details
        list servers request.
        """
        return {
            'name': self.server_name,
            'links': self.links_json(absolutize_url),
            'id': self.server_id
        }

    def detail_json(self, absolutize_url):
        """
        Long-form JSON-serializable object representation of this server, as
        returned by either a GET on this individual server or a member in the
        list returned by the list-details request.
        """
        template = self.static_defaults.copy()
        tenant_id = self.collection.tenant_id
        template.update({
            "id": self.server_id,
            "OS-DCF:diskConfig": self.disk_config,
            "OS-EXT-STS:vm_state": self.status,
            "addresses": self.addresses_json(),
            "created": seconds_to_timestamp(self.creation_time),
            "updated": seconds_to_timestamp(self.update_time),
            "flavor": {
                "id": self.flavor_ref,
                "links": [{
                    "href": absolutize_url(
                        "{0}/flavors/{1}".format(tenant_id, self.flavor_ref)),
                    "rel": "bookmark"}],
            },
            "image": {
                "id": self.image_ref,
                "links": [{
                    "href": absolutize_url("{0}/images/{1}".format(
                        tenant_id, self.flavor_ref)),
                    "rel": "bookmark"
                }]
            }
            if self.image_ref is not None else '',
            "links": self.links_json(absolutize_url),
            "key_name": self.key_name,
            "metadata": self.metadata,
            "name": self.server_name,
            "tenant_id": tenant_id,
            "status": self.status
        })
        return template

    def creation_response_json(self, absolutize_url):
        """
        A JSON-serializable object returned for the initial creation of this
        server.
        """
        return {
            'server': {
                "OS-DCF:diskConfig": self.disk_config,
                "id": self.server_id,
                "links": self.links_json(absolutize_url),
                "adminPass": self.admin_password,
            }
        }

    def set_metadata(self, metadata):
        """
        Replace all metadata with given metadata
        """
        self.metadata = metadata
        self.update_time = self.collection.clock.seconds()

    def set_metadata_item(self, key, value):
        """
        Set a metadata item on the server.

        All the response messages have been verified as of 2015-04-23 against
        Rackspace Nova.
        """
        if key not in self.metadata:
            if len(self.metadata) == self.max_metadata_items:
                raise LimitError(nova_message=(
                    "Maximum number of metadata items exceeds {0}"
                    .format(self.max_metadata_items)))

        if not isinstance(value, string_types):
            raise BadRequestError(nova_message=(
                "Invalid metadata: The input is not a string or unicode"))

        self.metadata[key] = value
        self.update_time = self.collection.clock.seconds()

    def update_status(self, status):
        """
        Update status on the server. This will also update the `update_time`
        of the server
        """
        self.status = status
        self.update_time = self.collection.clock.seconds()

    @classmethod
    def validate_metadata(cls, metadata, max_metadata_items=40):
        """
        Validate the given metadata - this is the complete metadata dict.

        All the response messages have been verified as of 2015-04-23 against
        Rackspace Nova.
        """
        if not isinstance(metadata, dict):
            raise BadRequestError(nova_message="Malformed request body")
        if len(metadata) > max_metadata_items:
            raise LimitError(nova_message=(
                "Maximum number of metadata items exceeds {0}"
                .format(max_metadata_items)))
        if not all(isinstance(v, string_types) for v in metadata.values()):
            raise BadRequestError(nova_message=(
                "Invalid metadata: The input is not a string or unicode"))

    @classmethod
    def from_creation_request_json(cls, collection, creation_json,
                                   ipsegment=lambda: randrange(255),
                                   max_metadata_items=40):
        """
        Create a :obj:`Server` from a JSON-serializable object that would be in
        the body of a create server request.
        """
        now = collection.clock.seconds()
        server_json = creation_json['server']
        disk_config = server_json.get('OS-DCF:diskConfig', None) or "AUTO"
        if disk_config not in ["AUTO", "MANUAL"]:
            raise BadRequestError(nova_message=(
                "OS-DCF:diskConfig must be either 'MANUAL' or 'AUTO'."))

        metadata = server_json.get("metadata") or {}
        cls.validate_metadata(metadata, max_metadata_items)

        while True:
            private_ip = IPv4Address(
                address="10.180.{0}.{1}".format(ipsegment(), ipsegment()))
            if private_ip not in [addr for server in collection.servers
                                  for addr in server.private_ips]:
                break

        self = cls(
            collection=collection,
            server_name=server_json['name'],
            server_id=('test-server{0}-id-{0}'
                       .format(str(randrange(9999999999)))),
            metadata=metadata,
            creation_time=now,
            update_time=now,
            private_ips=[private_ip],
            public_ips=[
                IPv4Address(address="198.101.241.{0}".format(ipsegment())),
                IPv6Address(address="2001:4800:780e:0510:d87b:9cbc:ff04:513a")
            ],
            key_name=None if 'key_name' not in server_json else server_json['key_name'],
            creation_request_json=creation_json,
            flavor_ref=server_json['flavorRef'],
            image_ref=server_json['imageRef'] or '',
            disk_config=disk_config,
            status="ACTIVE",
            admin_password=random_string(12),
            max_metadata_items=max_metadata_items
        )
        collection.servers.append(self)
        return self


@attributes(["address"])
class IPv4Address(object):
    """
    An IPv4 address for a server.
    """

    def json(self):
        """
        A JSON-serializable representation of this address.
        """
        return {"addr": self.address, "version": 4}


@attributes(["address"])
class IPv6Address(object):
    """
    An IPv6 address for a server.
    """

    def json(self):
        """
        A JSON-serializable representation of this address.
        """
        return {"addr": self.address, "version": 6}


server_creation = EventDescription()


@server_creation.declare_criterion("server_name")
def server_name_criterion(value):
    """
    Return a Criterion which matches the given regular expression string
    against the ``"server_name"`` attribute.
    """
    return Criterion(name='server_name', predicate=regexp_predicate(value))


@server_creation.declare_criterion("metadata")
def metadata_criterion(value):
    """
    Return a Criterion which matches against metadata.

    :param value: a dictionary, mapping a regular expression of a metadata key
        to a regular expression describing a metadata value.
    :type value: dict mapping unicode to unicode
    """
    def predicate(attribute):
        for k, v in value.items():
            if not re.compile(v).match(attribute.get(k, "")):
                return False
        return True
    return Criterion(name='metadata', predicate=predicate)


@server_creation.declare_default_behavior
def default_create_behavior(collection, http, json, absolutize_url,
                            ipsegment=lambda: randrange(255), hook=None):
    """
    Default behavior in response to a server creation.

    :param absolutize_url: A 1-argument function that takes a string and
        returns a string, where the input is the list of segments identifying a
        particular object within the compute service's URL hierarchy within a
        region, and the output is an absolute URL that identifies that same
        object.  Note that the region's URL hierarchy begins before the version
        identifier, because bookmark links omit the version identifier and go
        straight to the tenant ID.  Be sure to include the 'v2' first if you
        are generating a versioned URL; the tenant ID itself should always be
        passed in as part of the input, either the second or first segment,
        depending on whether the version is included or not respectively.

        Note that this is passed in on every request so that servers do not
        retain a memory of their full URLs internally, and therefore you may
        access Mimic under different hostnames and it will give you URLs
        appropriate to how you accessed it every time.  This is intentionally
        to support the use-case of running tests against your local dev machine
        as 'localhost' and then showing someone else the state that things are
        in when they will have to access your machine under a different
        hostname and therefore a different URI.

    :param ipsegment: A hook provided for IP generation so the IP addresses in
        tests are deterministic; normally a random number between 0 and 255.
    :param callable hook: a 1-argument callable which, if specified, will be
        invoked with the :obj:`Server` object after creating it, but before
        generating the response.  This allows for invoking the default behavior
        with a small tweak to alter the server's state in some way.
    """
    new_server = Server.from_creation_request_json(collection, json, ipsegment)
    if hook is not None:
        hook(new_server)
    response = new_server.creation_response_json(absolutize_url)
    http.setResponseCode(ACCEPTED)
    return dumps(response)


def default_with_hook(function):
    """
    A convenience decorator to make it easy to write a slightly-customized
    version of :obj:`default_create_behavior`.

    :param Server function: a 1-argument function taking a :obj:`Server` and
        returning Nothing.

    :return: a creation behavior, i.e. a function with the same signature as
             :obj:`default_create_behavior`, which does the default behavior of
             creating a server, adding it to the collection, and returning a
             successful ``ACCEPTED`` response, but with the server's state
             first modified by whatever the input ``function`` does.
    """
    def hooked(collection, http, json, absolutize_url):
        return default_create_behavior(collection, http, json, absolutize_url,
                                       hook=function)
    return hooked


def _get_failure_behavior(parameters, create=False):
    """
    Helper function to produce a failure to create function.  Either creating
    the server or not.

    Takes three parameters:

    ``"code"``, an integer describing the HTTP response code, and
    ``"message"``, a string describing a textual message.
    ``"type"``, a string representing what type of error message it is

    If ``type`` is "string", the message is just returned as the string body.
    Otherwise, the following JSON body will be synthesized (as per the
    canonical Nova error format):

    ```
    {
        <type>: {
            "message": <message>,
            "code": <code>
        }
    }

    The default type is computeFault, the default code is 500, and the default
    message is "The server has either erred or is incapable of performing the
    requested operation".
    """
    status_code = parameters.get("code", 500)
    failure_type = parameters.get("type", "computeFault")
    failure_message = parameters.get(
        "message",
        ("The server has either erred or is incapable of performing the "
         "requested operation"))

    if failure_type == "string":
        fail_body = failure_message
    else:
        fail_body = dumps({
            failure_type: {
                "message": failure_message,
                "code": status_code
            }
        })

    def _fail(collection, http, json, absolutize_url):
        if create:
            Server.from_creation_request_json(
                collection, json, lambda: randrange(255))

        http.setResponseCode(status_code)
        return fail_body
    return _fail


@server_creation.declare_behavior_creator("fail")
def create_fail_behavior(parameters):
    """
    Create a failing behavior for server creation.

    Takes three parameters:

    ``"code"``, an integer describing the HTTP response code, and
    ``"message"``, a string describing a textual message.
    ``"type"``, a string representing what type of error message it is

    If ``type`` is "string", the message is just returned as the string body.
    Otherwise, the following JSON body will be synthesized (as per the
    canonical Nova error format):

    ```
    {
        <type>: {
            "message": <message>,
            "code": <code>
        }
    }

    The default type is computeFault, the default code is 500, and the default
    message is "The server has either erred or is incapable of performing the
    requested operation".
    """
    return _get_failure_behavior(parameters)


@server_creation.declare_behavior_creator("false-negative")
def create_success_report_failure_behavior(parameters):
    """
    Create a behavior that reports failure, but actually succeeds, for server
    creation.

    Takes three parameters:

    ``"code"``, an integer describing the HTTP response code, and
    ``"message"``, a string describing a textual message.
    ``"type"``, a string representing what type of error message it is

    If ``type`` is "string", the message is just returned as the string body.
    Otherwise, the following JSON body will be synthesized (as per the
    canonical Nova error format):

    ```
    {
        <type>: {
            "message": <message>,
            "code": <code>
        }
    }

    The default type is computeFault, the default code is 500, and the default
    message is "The server has either erred or is incapable of performing the
    requested operation".
    """
    return _get_failure_behavior(parameters, create=True)


@server_creation.declare_behavior_creator("build")
def create_building_behavior(parameters):
    """
    Create a "build" behavior for server creation.

    Puts the server into the "BUILD" status immediately, transitioning it to
    "ACTIVE" after a requested amount of time.

    Takes one parameter:

    ``"duration"`` which is a Number, the duration of the build process in
    seconds.
    """
    duration = parameters["duration"]

    @default_with_hook
    def set_building(server):
        server.update_status(u"BUILD")
        server.collection.clock.callLater(
            duration,
            server.update_status,
            u"ACTIVE")
    return set_building


@server_creation.declare_behavior_creator("error")
def create_error_status_behavior(parameters=None):
    """
    Create an "error" behavior for server creation.

    The created server will go into the ``"ERROR"`` state immediately.

    Takes no parameters.
    """
    @default_with_hook
    def set_error(server):
        server.update_status(u"ERROR")
    return set_error


@server_creation.declare_behavior_creator("active-then-error")
def active_then_error(parameters):
    """
    Sometimes, a server goes into "active", but later (for unknown reasons)
    goes into "error"; presumably due to a hardware failure or similar
    operational issue.

    Takes one parameter:

    ``"duration"`` which is a Number, the duration of the time spent in the
    ``ACTIVE`` state.
    """
    duration = parameters["duration"]

    @default_with_hook
    def fail_later(server):
        server.update_status(u"ACTIVE")
        server.collection.clock.callLater(
            duration,
            server.update_status,
            u"ERROR")
    return fail_later


def metadata_to_creation_behavior(metadata):
    """
    Examine the metadata given to a server creation request, and return a
    behavior based on the values present there.
    """
    if 'create_server_failure' in metadata:
        return create_fail_behavior(loads(metadata['create_server_failure']))
    if 'server_building' in metadata:
        return create_building_behavior(
            {"duration": float(metadata['server_building'])}
        )
    if 'server_error' in metadata:
        return create_error_status_behavior()
    return None


@attributes(
    ["tenant_id", "region_name", "clock",
     Attribute("servers", default_factory=list),
     Attribute("image_store", default_factory=list),
     Attribute(
         "behavior_registry_collection",
         default_factory=lambda: BehaviorRegistryCollection())]
)
class RegionalServerCollection(object):
    """
    A collection of servers, in a given region, for a given tenant.
    """

    def server_by_id(self, server_id):
        """
        Retrieve a :obj:`Server` object by its ID.
        """
        for server in self.servers:
            if server.server_id == server_id and server.status != u"DELETED":
                return server

    def request_creation(self, creation_http_request, creation_json,
                         absolutize_url):
        """
        Request that a server be created.
        """
        metadata = creation_json.get('server', {}).get('metadata') or {}
        behavior = metadata_to_creation_behavior(metadata)
        if behavior is None:
            registry = self.behavior_registry_collection.registry_by_event(
                server_creation)
            behavior = registry.behavior_for_attributes({
                "tenant_id": self.tenant_id,
                "server_name": creation_json["server"]["name"],
                "metadata": creation_json["server"].get("metadata", {})
            })
        return behavior(self, creation_http_request, creation_json,
                        absolutize_url)

    def request_read(self, http_get_request, server_id, absolutize_url):
        """
        Request the information / details for an individual server.

        Not found response verified against Rackspace Cloud Servers as of
        2015-04-30.
        """
        server = self.server_by_id(server_id)
        if server is None:
            return dumps(not_found("Instance could not be found",
                                   http_get_request))
        return dumps({"server": server.detail_json(absolutize_url)})

    def request_ips(self, http_get_ips_request, server_id):
        """
        Request the addresses JSON for a specific server.

        Not found response verified against Rackspace Cloud Servers as of
        2015-04-30.
        """
        http_get_ips_request.setResponseCode(200)
        server = self.server_by_id(server_id)
        if server is None:
            return dumps(not_found("Instance does not exist",
                                   http_get_ips_request))
        return dumps({"addresses": server.addresses_json()})

    def request_list(self, http_get_request, include_details, absolutize_url,
                     name=u"", limit=None, marker=None, changes_since=None):
        """
        Request the list JSON for all servers.

        :param str changes_since: ISO8601 formatted datetime. Based on
            http://docs.rackspace.com/servers/api/v2/cs-devguide/content/ChangesSince.html

        Note: only supports filtering by name right now, but will need to
        support more going forward.

        Pagination behavior verified against Rackspace Nova as of 2015-04-29.
        """
        to_be_listed = self.servers

        if changes_since is not None:
            since = timestamp_to_seconds(changes_since)
            to_be_listed = filter(lambda s: s.update_time >= since, to_be_listed)

        # marker can be passed without limit, in which case the whole server
        # list, after the server that matches the marker, is returned
        if marker is not None:
            last_seen = [i for i, server in enumerate(to_be_listed)
                         if server.server_id == marker]
            if not last_seen:
                # Error response and body verified against Rackspace Nova as
                # of 2015-04-29
                return dumps(bad_request(
                    "marker [{0}] not found".format(marker),
                    http_get_request))
            else:
                last_seen = last_seen[0]
                to_be_listed = to_be_listed[last_seen + 1:]

        # A valid marker is an ID in the entire server list.  It does not
        # have to be for a server that matches the given name.
        to_be_listed = [server for server in to_be_listed
                        if name in server.server_name]

        if changes_since is None:
            to_be_listed = filter(lambda s: s.status != u"DELETED", to_be_listed)

        if limit is not None:
            try:
                limit = int(limit)
            except ValueError:
                return dumps(bad_request("limit param must be an integer",
                                         http_get_request))
            if limit < 0:
                return dumps(bad_request("limit param must be positive",
                                         http_get_request))

            to_be_listed = to_be_listed[:limit]

        result = {
            "servers": [
                server.brief_json(absolutize_url) if not include_details
                else server.detail_json(absolutize_url)
                for server in to_be_listed
            ]
        }

        # A server links blob is included only if limit is passed.  If
        # only the marker was provided, no server links blob is included.
        # Note that if limit=0, an empty server list is returned and no
        # server link blob is returned.
        if limit and len(to_be_listed) >= limit:
            query_params = {'limit': limit}
            query_params['marker'] = to_be_listed[-1].server_id
            if name:
                query_params['name'] = name

            path = "v2/{0}/servers{1}?{2}".format(
                self.tenant_id,
                "/detail" if include_details else "",
                urlencode(query_params))
            result["servers_links"] = [{
                "href": absolutize_url(path),
                "rel": "next"
            }]

        return dumps(result)

    def request_delete(self, http_delete_request, server_id):
        """
        Delete a server with the given ID.

        Not found response verified against Rackspace Cloud Servers as of
        2015-04-30.
        """
        server = self.server_by_id(server_id)
        if server is None:
            return dumps(not_found("Instance could not be found",
                                   http_delete_request))
        if 'delete_server_failure' in server.metadata:
            srvfail = loads(server.metadata['delete_server_failure'])
            if srvfail['times']:
                srvfail['times'] -= 1
                server.metadata['delete_server_failure'] = dumps(srvfail)
                http_delete_request.setResponseCode(500)
                return b''
        http_delete_request.setResponseCode(204)
        server.update_status(u"DELETED")
        return b''

    def request_action(self, http_action_request, server_id, absolutize_url):
        """
        Perform the requested action on the provided server
        """
        server = self.server_by_id(server_id)
        if server is None:
            return dumps(not_found("Instance " + server_id + " could not be found",
                                   http_action_request))
        action_json = loads(http_action_request.content.read())
        if 'resize' in action_json:
            flavor = action_json['resize'].get('flavorRef')
            if not flavor:
                return dumps(bad_request("Resize requests require 'flavorRef' attribute",
                                         http_action_request))

            server.status = 'VERIFY_RESIZE'
            server.oldFlavor = server.flavor_ref
            server.flavor_ref = flavor
            http_action_request.setResponseCode(202)
            return b''

        elif 'confirmResize' in action_json or 'revertResize' in action_json:
            if server.status == 'VERIFY_RESIZE' and 'confirmResize' in action_json:
                server.status = 'ACTIVE'
                http_action_request.setResponseCode(204)
                return b''
            elif server.status == 'VERIFY_RESIZE' and 'revertResize' in action_json:
                server.status = 'ACTIVE'
                server.flavor_ref = server.oldFlavor
                http_action_request.setResponseCode(202)
                return b''
            else:
                return dumps(conflicting("Cannot '" + action_json.keys()[0] + "' instance " + server_id +
                                         " while it is in vm_state active", http_action_request))
        elif 'rescue' in action_json:
            if server.status != 'ACTIVE':
                return dumps(conflicting("Cannot 'rescue' instance " + server_id +
                                         " while it is in task state other than active",
                                         http_action_request))
            else:
                server.status = 'RESCUE'
                http_action_request.setResponseCode(200)
                password = random_string(12)
                return dumps({"adminPass": password})

        elif 'unrescue' in action_json:
            if server.status == 'RESCUE':
                server.status = 'ACTIVE'
                http_action_request.setResponseCode(200)
                return b''
            else:
                return dumps(conflicting("Cannot 'unrescue' instance " + server_id +
                                         " while it is in task state other than rescue",
                                         http_action_request))

        elif 'reboot' in action_json:
            reboot_type = action_json['reboot'].get('type')
            if not reboot_type:
                return dumps(bad_request("Missing argument 'type' for reboot",
                                         http_action_request))
            if reboot_type == 'HARD':
                server.status = 'HARD_REBOOT'
                http_action_request.setResponseCode(202)
                server.collection.clock.callLater(
                    6.0,
                    server.update_status,
                    u"ACTIVE")
                return b''
            elif reboot_type == 'SOFT':
                server.status = 'REBOOT'
                http_action_request.setResponseCode(202)
                server.collection.clock.callLater(
                    3.0,
                    server.update_status,
                    u"ACTIVE")
                return b''
            else:
                return dumps(bad_request("Argument 'type' for reboot is not HARD or SOFT",
                                         http_action_request))

        elif 'changePassword' in action_json:
            password = action_json['changePassword'].get('adminPass')
            if not password:
                return dumps(bad_request("No adminPass was specified",
                                         http_action_request))
            if server.status == 'ACTIVE':
                http_action_request.setResponseCode(202)
                return b''
            else:
                return dumps(conflicting("Cannot 'changePassword' instance " + server_id +
                                         " while it is in task state other than active",
                                         http_action_request))

        elif 'rebuild' in action_json:
            image_ref = action_json['rebuild'].get('imageRef')
            if not image_ref:
                return dumps(bad_request("Could not parse imageRef from request.", http_action_request))
            if server.status == 'ACTIVE':
                server.image_ref = image_ref
                server.status = 'REBUILD'
                http_action_request.setResponseCode(202)
                server.collection.clock.callLater(
                    5.0,
                    server.update_status,
                    u"ACTIVE")
                server_details = server.detail_json(absolutize_url)
                server_details['adminPass'] = 'password'
                return dumps({"server": server_details})
            else:
                return dumps(conflicting("Cannot 'rebuild' instance " + server_id +
                                         " while it is in task state other than active",
                                         http_action_request))

        else:
            return dumps(bad_request("There is no such action currently supported", http_action_request))


@attributes(["tenant_id", "clock",
             Attribute("regional_collections", default_factory=dict)])
class GlobalServerCollections(object):
    """
    A :obj:`GlobalServerCollections` is a set of all the
    :obj:`RegionalServerCollection` objects owned by a given tenant.  In other
    words, all the objects that a single tenant owns globally in a Nova
    service.
    """

    def collection_for_region(self, region_name):
        """
        Get a :obj:`RegionalServerCollection` for the region identified by the
        given name.
        """
        if region_name not in self.regional_collections:
            self.regional_collections[region_name] = (
                RegionalServerCollection(tenant_id=self.tenant_id,
                                         region_name=region_name,
                                         clock=self.clock)
            )
        return self.regional_collections[region_name]



# -*- test-case-name: mimic.test.test_loadbalancer -*-
"""
Canned response for add/get/list/delete load balancers and
add/get/delete/list nodes
"""
from random import randrange


def load_balancer_example(lb_info, lb_id, status,
                          current_time):
    """
    Create load balancer response example
    """
    lb_example = {"name": lb_info["name"],
                  "id": lb_id,
                  "protocol": lb_info["protocol"],
                  "port": lb_info.get("port", 80),
                  "algorithm": lb_info.get("algorithm") or "RANDOM",
                  "status": status,
                  "cluster": {"name": "test-cluster"},
                  "timeout": lb_info.get("timeout", 30),
                  "created": {"time": current_time},
                  "virtualIps": [{"address": "127.0.0.1",
                                 "id": 1111, "type": "PUBLIC", "ipVersion": "IPV4"},
                                 {"address": "0000:0000:0000:0000:1111:111b:0000:0000",
                                  "id": 1111,
                                  "type": "PUBLIC",
                                  "ipVersion": "IPV6"}],
                  "sourceAddresses": {"ipv6Public": "0000:0001:0002::00/00",
                                      "ipv4Servicenet": "127.0.0.1",
                                      "ipv4Public": "127.0.0.1"},
                  "httpsRedirect": lb_info.get("httpsRedirect", False),
                  "updated": {"time": current_time},
                  "halfClosed": lb_info.get("halfClosed", False),
                  "connectionLogging": lb_info.get("connectionLogging", {"enabled": False}),
                  "contentCaching": {"enabled": False}}
    if lb_info.get("metadata"):
        lb_example.update({"metadata": _format_meta(lb_info["metadata"])})
    return lb_example


def _format_meta(metadata_list):
    """
    creates metadata with 'id' as a key
    """
    meta = []
    for each in metadata_list:
        each.update({"id": randrange(999)})
        meta.append(each)
    return meta




from json import dumps

from twisted.trial.unittest import SynchronousTestCase
from twisted.internet.task import Clock

import treq

from mimic.rest.swift_api import SwiftMock
from mimic.resource import MimicRoot
from mimic.core import MimicCore
from mimic.rest.swift_api import normal_tenant_id_to_crazy_mosso_id
from mimic.test.helpers import request


class SwiftTests(SynchronousTestCase):
    """
    tests for swift API
    """

    def createSwiftService(self, rackspace_flavor=True):
        """
        Set up to create the requests
        """
        self.core = MimicCore(Clock(), [SwiftMock(rackspace_flavor)])
        self.root = MimicRoot(self.core).app.resource()
        self.response = request(
            self, self.root, "POST", "/identity/v2.0/tokens",
            dumps({
                "auth": {
                    "passwordCredentials": {
                        "username": "test1",
                        "password": "test1password",
                    },
                    # TODO: should this really be 'tenantId'?
                    "tenantName": "fun_tenant",
                }
            })
        )
        self.auth_response = self.successResultOf(self.response)
        self.json_body = self.successResultOf(
            treq.json_content(self.auth_response))

    def test_service_catalog(self):
        """
        When provided with a :obj:`SwiftMock`, :obj:`MimicCore` yields a
        service catalog containing a swift endpoint.
        """
        self.createSwiftService()
        self.assertEqual(self.auth_response.code, 200)
        self.assertTrue(self.json_body)
        sample_entry = self.json_body['access']['serviceCatalog'][0]
        self.assertEqual(sample_entry['type'], u'object-store')
        sample_endpoint = sample_entry['endpoints'][0]
        self.assertEqual(
            sample_endpoint['tenantId'],
            normal_tenant_id_to_crazy_mosso_id("fun_tenant")
        )
        self.assertEqual(sample_endpoint['region'], 'ORD')
        self.assertEqual(len(self.json_body['access']['serviceCatalog']), 1)

    def create_one_container(self, expected_code):
        """
        Create one container and assert its code is the given expected status.
        """
        uri = (self.json_body['access']['serviceCatalog'][0]['endpoints'][0]
               ['publicURL'] + '/testcontainer')
        create_container = request(self, self.root, "PUT", uri)
        create_container_response = self.successResultOf(create_container)
        self.assertEqual(create_container_response.code, expected_code)
        self.assertEqual(
            self.successResultOf(treq.content(create_container_response)),
            b"",
        )

    def test_create_container(self):
        """
        Test to verify create container using :obj:`SwiftMock`
        """
        self.createSwiftService()
        self.create_one_container(201)

    def test_create_twice(self):
        """
        Creating a container twice results in an ACCEPTED status code.
        """
        self.createSwiftService()
        self.create_one_container(201)
        self.create_one_container(202)
        self.create_one_container(202)

    def test_get_container(self):
        """
        Creating a container and immediately retrieving it yields an empty list
        (since there are no objects) and several headers indicating that no
        objects are in the container and they consume no space.
        """
        self.createSwiftService()
        # create a container
        uri = (self.json_body['access']['serviceCatalog'][0]['endpoints'][0]
               ['publicURL'] + '/testcontainer')
        create_container = request(self, self.root, "PUT", uri)
        self.successResultOf(create_container)
        container_response = self.successResultOf(
            request(self, self.root, "GET", uri)
        )
        self.assertEqual(container_response.code, 200)
        container_contents = self.successResultOf(
            treq.json_content(container_response)
        )
        self.assertEqual(container_contents, [])
        self.assertEqual(
            container_response.headers.getRawHeaders(
                "X-Container-Object-Count")[0], "0"
        )
        self.assertEqual(
            container_response.headers.getRawHeaders(
                "X-Container-Bytes-Used")[0], "0"
        )

    def test_get_no_container(self):
        """
        GETing a container that has not been created results in a 404.
        """
        self.createSwiftService()
        # create a container
        uri = (self.json_body['access']['serviceCatalog'][0]['endpoints'][0]
               ['publicURL'] + '/testcontainer')
        container_response = self.successResultOf(
            request(self, self.root, "GET", uri)
        )
        self.assertEqual(container_response.code, 404)
        self.assertEqual(
            container_response.headers.getRawHeaders(
                "X-Container-Object-Count"), None
        )
        self.assertEqual(
            container_response.headers.getRawHeaders(
                "X-Container-Bytes-Used"), None
        )

    def test_put_object(self):
        """
        PUTting an object into a container causes the container to list that
        object.
        """
        self.createSwiftService()
        # create a container
        uri = (self.json_body['access']['serviceCatalog'][0]['endpoints'][0]
               ['publicURL'] + '/testcontainer')
        create_container = request(self, self.root, "PUT", uri)
        self.successResultOf(create_container)
        BODY = b'some bytes'
        object_uri = uri + "/" + "testobject"
        object_response = request(self, self.root,
                                  "PUT", object_uri,
                                  headers={"content-type": ["text/plain"]},
                                  body=BODY)
        self.assertEqual(self.successResultOf(object_response).code,
                         201)
        container_response = self.successResultOf(
            request(self, self.root, "GET", uri)
        )
        self.assertEqual(container_response.code, 200)
        container_contents = self.successResultOf(
            treq.json_content(container_response)
        )
        self.assertEqual(len(container_contents), 1)
        self.assertEqual(container_contents[0]['name'], "testobject")
        self.assertEqual(container_contents[0]['content_type'], "text/plain")
        self.assertEqual(container_contents[0]['bytes'], len(BODY))
        object_response = self.successResultOf(
            request(self, self.root, "GET", object_uri)
        )
        self.assertEqual(object_response.code, 200)
        object_body = self.successResultOf(treq.content(object_response))
        self.assertEquals(object_body, BODY)

    def test_openstack_ids(self):
        """
        Non-Rackspace implementations of Swift just use the same tenant ID as
        other services in the catalog.

        (Note that this is not exposed by configuration yet, see
        U{https://github.com/rackerlabs/mimic/issues/85})
        """
        self.createSwiftService(False)
        url = (self.json_body['access']['serviceCatalog'][0]
               ['endpoints'][0]['publicURL'])
        self.assertIn("/fun_tenant", url)
        self.assertNotIn("/MossoCloudFS_", url)
